#!/usr/bin/env python3
"""A script runner entrypoint for DP3 scripts."""
import subprocess
import sys
from argparse import ArgumentParser
from importlib.util import find_spec
from os import chmod
from pathlib import Path


class C:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


def red(text: str):
    return f"{C.FAIL}{text}{C.ENDC}"


def yellow(text: str):
    return f"{C.WARNING}{text}{C.ENDC}"


def bold(text: str):
    return f"{C.BOLD}{text}{C.ENDC}"


def try_pandas_import():
    if find_spec("pandas") is None:
        print(
            yellow("WARNING: pandas not found. Some scripts may not work as expected.") + "\n"
            "You can install it with other script dependencies using:\n"
            "  pip install dp-cubed[scripts]\n",
        )


def run():
    parser = ArgumentParser(
        description=f"A script runner entrypoint for {bold('DP3')} scripts.\n"
        "You can enter a script name with arguments, or any of the available commands.\n"
        "Available commands: ls\n",
    )
    parser.add_argument(
        "command",
        help="The command or script to run. Available commands: ls",
    )
    args = parser.parse_args(sys.argv[1:2])
    file_path = Path(__file__)
    script_dir = file_path.parent

    if args.command == "ls":
        for item in sorted(script_dir.iterdir(), key=lambda x: x.name):
            if item.name.startswith("__") or item.name == file_path.name:
                continue
            print(f"{item.name}")
        return
    else:
        script_path = script_dir / args.command
        if not script_path.exists():
            print(red(f"ERROR: Script '{args.command}' not found."))
            return

        try_pandas_import()

        chmod(script_path, 0o775)
        subprocess.run([script_path, *sys.argv[2:]], check=False)



import os
import sys
from enum import Enum

from pydantic import BaseModel, ValidationError, field_validator

from dp3.api.internal.dp_logger import DPLogger
from dp3.common.config import ModelSpec, read_config_dir
from dp3.database.database import EntityDatabase
from dp3.history_management.telemetry import TelemetryReader
from dp3.task_processing.task_queue import TaskQueueWriter

DATAPOINTS_INGESTION_URL_PATH = "/datapoints"


class ConfigEnv(BaseModel):
    """Configuration environment variables container"""

    APP_NAME: str
    CONF_DIR: str
    ROOT_PATH: str = ""

    @field_validator("CONF_DIR")
    @classmethod
    def validate(cls, v):
        # Try to open config
        try:
            config = read_config_dir(v, recursive=True)
        except OSError as e:
            raise ValueError(f"Cannot open: {v}") from e

        assert "db_entities" in config, "Config for 'db_entities' is missing"
        assert "database" in config, "Config for 'database' is missing"

        # This may raise ValueError too
        ModelSpec(config.get("db_entities"))

        return v


try:
    # Validate and parse environmental variables
    conf_env = ConfigEnv.parse_obj(os.environ)
except ValidationError as e:
    config_error = any("CONF_DIR" in x["loc"] and len(x["loc"]) > 1 for x in e.errors())
    env_error = any(len(x["loc"]) == 1 for x in e.errors())
    print(
        ("Invalid or missing environmental variables" if env_error else "")
        + (" && " if env_error and config_error else "")
        + ("Invalid model specification (check entity config)" if config_error else "")
        + ":",
        file=sys.stderr,
    )

    error_message_no_first_line = str(e).split("\n", 1)[-1]
    print(error_message_no_first_line, file=sys.stderr)

    sys.exit(1)

# Load configuration, entity and attribute specification and connect to DP3 message broker.
CONFIG = read_config_dir(conf_env.CONF_DIR, recursive=True)
EnabledModules = Enum(
    "EnabledModules", {module: module for module in CONFIG.get("processing_core.enabled_modules")}
)
MODEL_SPEC = ModelSpec(CONFIG.get("db_entities"))
DB = EntityDatabase(CONFIG, MODEL_SPEC, CONFIG.get("processing_core.worker_processes"))
TASK_WRITER = TaskQueueWriter(
    conf_env.APP_NAME,
    CONFIG.get("processing_core.worker_processes"),
    CONFIG.get("processing_core.msg_broker"),
)
CONTROL_WRITER = TaskQueueWriter(
    conf_env.APP_NAME,
    CONFIG.get("processing_core.worker_processes"),
    CONFIG.get("processing_core.msg_broker"),
    exchange=f"{conf_env.APP_NAME}-control-exchange",
)
DP_LOGGER = DPLogger(CONFIG.get("api.datapoint_logger"))
ROOT_PATH = conf_env.ROOT_PATH
TELEMETRY_READER = TelemetryReader(DB)



from typing import Literal, Union

from pydantic import BaseModel, Extra, Field

from dp3.common.types import ParsedTimedelta


class SpecModel(BaseModel, extra=Extra.forbid): ...


class ImmortalLifetime(SpecModel):
    """Immortal lifetime specification.

    The entity is never deleted.
    """

    type: Literal["immortal"]


class TimeToLiveLifetime(SpecModel):
    """TTL lifetime specification.

    The entity is deleted after all of its TTL tokens expire.
    TTL tokens can be attached to new data datapoints on per-attribute basis
    (see [`AttrSpecGeneric`][dp3.common.attrspec.AttrSpecGeneric]),
    set to mirror the lifetime of the data (`mirror_data`),
    or sent explicitly using the API (see `/entity/{etype}/{eid}/ttl`).

    Attributes:
        on_create: The base lifetime of an entity.
        mirror_data: If `True` (default), the lifetime of the entity is extended by the
            `max_age` of the incoming observations and timeseries data-points.
    """

    type: Literal["ttl"]
    on_create: ParsedTimedelta
    mirror_data: bool = True


class WeakLifetime(SpecModel):
    """Weak entity lifetime specification"""

    type: Literal["weak"]


class EntitySpec(SpecModel):
    """Entity specification

    This class represents specification of an entity type (e.g. ip, asn, ...)
    """

    id: str
    name: str
    snapshot: bool
    lifetime: Union[ImmortalLifetime, TimeToLiveLifetime, WeakLifetime] = Field(
        default_factory=lambda: ImmortalLifetime(type="immortal"), discriminator="type"
    )

    description: str = ""



import logging
import pathlib

from dp3.common.datapoint import DataPointBase


class DPLogger:
    """Datapoint logger

    Logs good/bad datapoints into file for further analysis.
    They are logged in JSON format.
    Bad datapoints are logged together with their error message.

    Logging may be disabled in `api.yml` configuration file:

    ```yml
    # ...
    datapoint_logger:
      good_log: false
      bad_log: false
    # ...
    ```

    """

    LOG_FORMATTER = logging.Formatter("%(asctime)s (%(src)s) | %(message)s")
    UNKNOWN_SRC_MSG = "UNKNOWN"

    def __init__(self, config: dict):
        if not config:
            config = {}

        good_log_file = config.get("good_log", False)
        bad_log_file = config.get("bad_log", False)

        # Setup loggers
        self._good_logger = self.setup_logger("GOOD", good_log_file)
        self._bad_logger = self.setup_logger("BAD", bad_log_file)

    def setup_logger(self, name: str, log_file: str):
        """Creates new logger instance with `log_file` as target"""
        # Create log handler
        if log_file:
            parent_path = pathlib.Path(log_file).parent
            if not parent_path.exists():
                raise FileNotFoundError(
                    f"The directory {parent_path} does not exist,"
                    " check the configured path or create the directory."
                )
            log_handler = logging.FileHandler(log_file)
            log_handler.setFormatter(self.LOG_FORMATTER)
        else:
            log_handler = logging.NullHandler()

        # Get logger instance
        logger = logging.getLogger(name)
        logger.addHandler(log_handler)
        logger.setLevel(logging.INFO)
        logger.propagate = False

        return logger

    def log_good(self, dps: list[DataPointBase], src: str = UNKNOWN_SRC_MSG):
        """Logs good datapoints

        Datapoints are logged one-by-one in processed form.
        Source should be IP address of incomping request.
        """
        for dp in dps:
            self._good_logger.info(dp.model_dump(), extra={"src": src})

    def log_bad(self, validation_error_msg: str, src: str = UNKNOWN_SRC_MSG):
        """Logs validation error message, which includes bad input datapoints

        Should be called for each individual error (JSON string is expected).
        Source should be IP address of incoming request.
        """
        self._bad_logger.info(validation_error_msg, extra={"src": src})



import os
import unittest
from copy import deepcopy

from dp3.common.config import HierarchicalDict, read_config_dir


class TestHierarchicalDict(unittest.TestCase):
    d1 = HierarchicalDict(
        {
            "a": "a",
            "b": {
                "b1": 1,
                "b2": [1, 2, 3],
            },
            "c": {
                "c1": {
                    "c1x": "qwer",
                    "c1y": "asdf",
                },
                "c2": 2,
            },
        }
    )

    d2 = HierarchicalDict(
        {
            "b": {
                "b1": 12345,
                "b3": "BBB",
                "b4": {"b4a": None},
            },
            "c": {
                "c2": {"c20": None},
            },
            "d": "x",
            "e": {},
        }
    )

    d3 = {
        "a": "A",
    }

    d4 = [("c", ["test"]), ("d", "y")]
    d5 = {"c": {"c1": {"c1z": None}}}
    d_empty = {}

    # Expected results
    d1d2 = HierarchicalDict(
        {
            "a": "a",
            "b": {
                "b1": 12345,
                "b2": [1, 2, 3],
                "b3": "BBB",
                "b4": {"b4a": None},
            },
            "c": {
                "c1": {
                    "c1x": "qwer",
                    "c1y": "asdf",
                },
                "c2": {"c20": None},
            },
            "d": "x",
            "e": {},
        }
    )

    d2d3d4 = HierarchicalDict(
        {
            "a": "A",
            "b": {
                "b1": 12345,
                "b3": "BBB",
                "b4": {"b4a": None},
            },
            "c": ["test"],
            "d": "y",
            "e": {},
        }
    )

    d1d5 = HierarchicalDict(
        {
            "a": "a",
            "b": {
                "b1": 1,
                "b2": [1, 2, 3],
            },
            "c": {
                "c1": {
                    "c1x": "qwer",
                    "c1y": "asdf",
                    "c1z": None,
                },
                "c2": 2,
            },
        }
    )

    def test_update(self):
        # test update of HierarchicalDict
        result = deepcopy(self.d1)
        result.update(self.d2)
        self.assertEqual(result, self.d1d2, "Update of d2 by d2 failed!")

        result = deepcopy(self.d2)
        result.update(self.d3)
        result.update(self.d4)
        self.assertEqual(result, self.d2d3d4, "Update of d2 by d3 and d4 failed!")

        result = deepcopy(self.d1)
        result.update(self.d5)
        self.assertEqual(result, self.d1d5, "Update of d1 by d5 failed!")

        result = deepcopy(self.d1)
        result.update(self.d_empty)
        result.update([])
        result.update(HierarchicalDict())
        self.assertEqual(result, self.d1, "Dict has changed after update by empty dict!")

    def test_get(self):
        # test get of HierarchicalDict
        self.assertEqual(self.d1.get("b.b1"), 1, "Dot notation test in d1 failed!")
        self.assertEqual(
            self.d1d5.get("c.c1"), self.d1d5["c"]["c1"], "Dot notation test in d1d5 failed!"
        )
        self.assertEqual(self.d2d3d4.get("b.b4.b4a"), None, "Dot notation in d2d3d4 failed!")
        self.assertEqual(self.d1d2.get("c.c1.c1x"), "qwer", "Dot notation in d1d2 failed!")


class TestReadConfig(unittest.TestCase):
    test_config_dir_recursive = {
        "file1": {
            "msg_broker": {
                "host": "localhost",
                "port": 5672,
                "virtual_host": "/",
                "username": "guest",
                "password": "guest",
            },
            "worker_processes": 1,
        },
        "file2": {"mongodb": {"host": "localhost", "port": 27017, "dbname": "nerd"}},
        "modules": {
            "module1": {
                "record_life_length": {
                    "warden": 14,
                    "misp": 180,
                    "highly_active": 14,
                    "long_active": 28,
                },
                "record_life_threshold": {"highly_active": 1000, "long_active": 30},
            },
            "module2": {
                "rate-limit": {"tokens-per-sec": 1, "bucket-size": 60, "redis": {"db-index": 1}}
            },
        },
    }

    test_config_dir_nonrecursive = {
        "file1": {
            "msg_broker": {
                "host": "localhost",
                "port": 5672,
                "virtual_host": "/",
                "username": "guest",
                "password": "guest",
            },
            "worker_processes": 1,
        },
        "file2": {"mongodb": {"host": "localhost", "port": 27017, "dbname": "nerd"}},
    }

    CONFIG_TEST_DIR = "config_test_files"

    def test_read_config_dir_recursive(self):
        loaded_config = read_config_dir(
            os.path.join(os.path.dirname(__file__), self.CONFIG_TEST_DIR), recursive=True
        )
        self.assertEqual(
            self.test_config_dir_recursive,
            loaded_config,
            f"Loaded config from {self.CONFIG_TEST_DIR} directory recursively did not match!",
        )

    def test_read_config_dir_nonrecursive(self):
        loaded_config = read_config_dir(
            os.path.join(os.path.dirname(__file__), self.CONFIG_TEST_DIR)
        )
        self.assertEqual(
            self.test_config_dir_nonrecursive,
            loaded_config,
            f"Loaded config from {self.CONFIG_TEST_DIR} directory nonrecursively did not match!",
        )



"""
# Dynamic Profile Processing Platform (DP³)

Platform directory structure:

- [Worker][dp3.worker] - The main worker process.

- [API][dp3.api] - The HTTP API implementation.

- [Binaries][dp3.bin] under `dp3.bin` - Executables for running the platform.

- [Common][dp3.common] - Common modules which are used throughout the platform.
    - [Config][dp3.common.config], [EntitySpec][dp3.common.entityspec] and
    [AttrSpec][dp3.common.attrspec] - Models for reading, validation and representing
    platform configuration of entities and their attributes.
    [datatype][dp3.common.datatype] is also used within this context.
    - [Scheduler][dp3.common.scheduler] - Allows modules to run callbacks at specified times
    - [Task][dp3.common.task.Task] - Model for a single task processed by the platform
    - [Utils][dp3.common.utils] - Auxiliary utility functions

- [Database.EntityDatabase][dp3.database.database] - A wrapper responsible for communication
with the database server.

- [HistoryManagement.HistoryManager][dp3.history_management.history_manager] - Module responsible
for managing history saved in database, currently to clean old data.

- [Snapshots][dp3.snapshots] - [SnapShooter][dp3.snapshots.snapshooter], a module responsible
for snapshot creation and running configured data correlation and fusion hooks,
and [Snapshot Hooks][dp3.snapshots.snapshot_hooks], which manage the registered hooks and their
dependencies on one another.

- [TaskProcessing][dp3.task_processing] - Module responsible for task
[distribution][dp3.task_processing.task_distributor],
[processing][dp3.task_processing.task_executor] and running configured
[hooks][dp3.task_processing.task_hooks]. Task distribution is possible due to the
[task queue][dp3.task_processing.task_queue].
"""



from time import time

from dp3.common.base_module import BaseModule
from dp3.common.callback_registrar import CallbackRegistrar
from dp3.common.config import PlatformConfig
from dp3.common.task import DataPointTask


class TestModule(BaseModule):
    def __init__(
        self, platform_config: PlatformConfig, module_config: dict, registrar: CallbackRegistrar
    ):
        super().__init__(platform_config, module_config, registrar)

        registrar.register_entity_hook(
            "on_entity_creation", hook=self.fill_test_attr_string, entity="test_entity_type"
        )

        registrar.register_correlation_hook(
            hook=self.processing_func_timestamp,
            entity_type="test_entity_type",
            depends_on=[["test_attr_string"]],
            may_change=[["test_attr_time"]],
        )

        registrar.register_correlation_hook(
            hook=self.processing_func_test_attrib,
            entity_type="test_entity_type",
            depends_on=[["test_attr_time"]],
            may_change=[["test_attr_int"]],
        )

    def load_config(self, config: PlatformConfig, module_config: dict) -> None:
        # just for testing purposes - as new value for test_attrib
        self.counter = module_config.get("init_value", 0)
        self.msg = module_config.get("msg", "Hello World!")

    def fill_test_attr_string(self, eid: str, task: DataPointTask) -> list[DataPointTask]:
        """receives eid and Task, may return new Tasks (including new DataPoints)"""
        return [
            DataPointTask(
                etype="test_entity_type",
                eid=eid,
                data_points=[
                    {
                        "etype": "test_entity_type",
                        "eid": eid,
                        "attr": "test_attr_string",
                        "src": "secondary/test_module",
                        "v": f"entity key {eid}",
                    }
                ],
            )
        ]

    def processing_func_timestamp(self, eid: str, record: dict):
        """
        Set current time to 'test_timestamp'
        Args:
            eid: entity identifier
            record: record with current values
        Returns:
            new update as Task.
        """
        self.log.info("Hello from TestModule - processing_func_timestamp")
        current_time = time()

        record["test_attr_time"] = current_time

    def processing_func_test_attrib(self, eid: str, record: dict):
        """
        Increase test_attrib's value by one.
        Args:
            eid: entity identificator
            record: record with current values
        Returns:
            new update as Task.
        """
        self.log.info("Hello from TestModule - processing_func_attrib")
        self.counter += 1

        record["test_attrib"] = self.counter



"""
auxiliary/utility functions and classes
"""

import datetime
import re
from collections.abc import Iterable, Iterator
from functools import partial
from itertools import islice
from typing import Union

# *** IP conversion functions ***
ipv4_re = re.compile(r"^([0-9]{1,3})\.([0-9]{1,3})\.([0-9]{1,3})\.([0-9]{1,3})$")


def entity_expired(utcnow: datetime, master_document: dict):
    """Check if entity is expired (all TTLs are in the past)"""
    if "#ttl" not in master_document:
        return False
    return all(ttl < utcnow for ttl in master_document["#ttl"].values())


def ipstr2int(s):
    res = ipv4_re.match(s)
    if res is None:
        raise ValueError(f"Invalid IPv4 format: {s!r}")
    a1, a2, a3, a4 = res.groups()
    # Check if octets are between 0 and 255 is omitted for better performance
    return int(a1) << 24 | int(a2) << 16 | int(a3) << 8 | int(a4)


def int2ipstr(i):
    return ".".join((str(i >> 24), str((i >> 16) & 0xFF), str((i >> 8) & 0xFF), str(i & 0xFF)))


# *** Time conversion ***
# Regex for RFC 3339 time format
timestamp_re = re.compile(
    r"^(\d{4})-(\d{2})-(\d{2})[Tt ](\d{2}):(\d{2}):(\d{2})(?:\.(\d+))?([Zz]|(?:[+-]\d{2}:\d{2}))?$"
)


def parse_rfc_time(time_str):
    """
    Parse time in RFC 3339 format and return it as naive datetime in UTC.

    Timezone specification is optional (UTC is assumed when none is specified).
    """
    res = timestamp_re.match(time_str)
    if res is not None:
        year, month, day, hour, minute, second = (int(n or 0) for n in res.group(*range(1, 7)))
        us_str = (res.group(7) or "0")[:6].ljust(6, "0")
        us = int(us_str)
        zonestr = res.group(8)
        zoneoffset = 0 if zonestr in (None, "z", "Z") else int(zonestr[:3]) * 60 + int(zonestr[4:6])
        zonediff = datetime.timedelta(minutes=zoneoffset)
        return datetime.datetime(year, month, day, hour, minute, second, us) - zonediff
    else:
        raise ValueError("Wrong timestamp format")


time_duration_pattern = re.compile(r"^\s*(\d+)([smhd])?$")


def parse_time_duration(duration_string: Union[str, int, datetime.timedelta]) -> datetime.timedelta:
    """
    Parse duration in format <num><s/m/h/d> (or just "0").

    Return datetime.timedelta
    """
    # if it's already timedelta, just return it unchanged
    if isinstance(duration_string, datetime.timedelta):
        return duration_string
    # if number is passed, consider it number of seconds
    if isinstance(duration_string, (int, float)):
        return datetime.timedelta(seconds=duration_string)

    d = 0
    h = 0
    m = 0
    s = 0

    if duration_string == "0":
        pass
    elif duration_string[-1] == "d":
        d = int(duration_string[:-1])
    elif duration_string[-1] == "h":
        h = int(duration_string[:-1])
    elif duration_string[-1] == "m":
        m = int(duration_string[:-1])
    elif duration_string[-1] == "s":
        s = int(duration_string[:-1])
    else:
        raise ValueError("Invalid time duration string")

    return datetime.timedelta(days=d, hours=h, minutes=m, seconds=s)


# *** object (de)serialization ***
# Functions that allow to (de)serialize some objects we need to pass for example via TaskQueue.
# Inspired by bson.json_util, but for different set of types (and much simpler).
def conv_to_json(obj):
    """Convert special types to JSON (use as "default" param of json.dumps)

    Supported types/objects:
    - datetime
    - timedelta
    """
    if isinstance(obj, datetime.datetime):
        if obj.tzinfo:
            raise NotImplementedError(
                "Can't serialize timezone-aware datetime object "
                "(DP3 policy is to use naive datetimes in UTC everywhere)"
            )
        return {"$datetime": obj.strftime("%Y-%m-%dT%H:%M:%S.%f")}
    if isinstance(obj, datetime.timedelta):
        return {"$timedelta": f"{obj.days},{obj.seconds},{obj.microseconds}"}
    raise TypeError(f"{repr(obj)}%r is not JSON serializable")


def conv_from_json(dct):
    """Convert special JSON keys created by conv_to_json back to Python objects
    (use as "object_hook" param of json.loads)

    Supported types/objects:
    - datetime
    - timedelta
    """
    if "$datetime" in dct:
        val = dct["$datetime"]
        return datetime.datetime.strptime(val, "%Y-%m-%dT%H:%M:%S.%f")
    if "$timedelta" in dct:
        days, seconds, microseconds = dct["$timedelta"].split(",")
        return datetime.timedelta(int(days), int(seconds), int(microseconds))
    return dct


def batched(iterable: Iterable, n: int) -> Iterator[list]:
    """Batch data into tuples of length n. The last batch may be shorter."""
    # batched('ABCDEFG', 3) --> ABC DEF G
    if n < 1:
        raise ValueError("n must be at least one")
    it = iter(iterable)
    while True:
        batch = list(islice(it, n))
        if not batch:
            break
        yield batch


# *** pretty print ***
def get_func_name(func_or_method):
    """Get name of function or method as pretty string."""
    if isinstance(func_or_method, partial):
        wrapper = "partial({name}, {args})"
        args = [str(a) for a in func_or_method.args]
        args.extend(f"{k}={v}" for k, v in func_or_method.keywords.items())
        args = ", ".join(args)
        func_or_method = func_or_method.func
    else:
        wrapper = "{name}{args}"
        args = ""

    try:
        fname = func_or_method.__func__.__qualname__
    except AttributeError:
        try:
            fname = func_or_method.__name__
        except AttributeError:
            fname = str(func_or_method)

    try:
        module = func_or_method.__module__
    except AttributeError:
        return fname
    return wrapper.format(name=f"{module}.{fname}", args=args)



"""
Module managing creation of snapshots, enabling data correlation and saving snapshots to DB.

- Snapshots are created periodically (user configurable period)

- When a snapshot is created, several things need to happen:
    - all registered timeseries processing modules must be called
      - this should result in `observations` or `plain` datapoints, which will be saved to db
        and forwarded in processing
    - current value must be computed for all observations
      - load relevant section of observation's history and perform configured history analysis.
        Result = plain values
    - load plain attributes saved in master collection
    - A record of described plain data makes a `profile`
    - Profile is additionally extended by related entities
    - Callbacks for data correlation and fusion should happen here
    - Save the complete results into database as snapshots
"""

import logging
from collections import defaultdict
from datetime import datetime
from typing import Any, Callable, Optional, Union

import pymongo.errors
from event_count_logger import DummyEventGroup
from pydantic import BaseModel

from dp3.common.attrspec import (
    AttrSpecObservations,
    AttrSpecType,
    AttrType,
    ObservationsHistoryParams,
)
from dp3.common.config import CronExpression, PlatformConfig
from dp3.common.scheduler import Scheduler
from dp3.common.task import DataPointTask, Snapshot, SnapshotMessageType, task_context
from dp3.common.types import EventGroupType
from dp3.common.utils import get_func_name
from dp3.database.database import EntityDatabase
from dp3.snapshots.snapshot_hooks import (
    SnapshotCorrelationHookContainer,
    SnapshotTimeseriesHookContainer,
)
from dp3.task_processing.task_queue import TaskQueueReader, TaskQueueWriter

DB_SEND_CHUNK = 100
RETRY_COUNT = 3


class SnapShooterConfig(BaseModel):
    creation_rate: CronExpression = CronExpression(minute="*/30")
    keep_empty: bool = True


class SnapShooter:
    """Class responsible for creating entity snapshots."""

    def __init__(
        self,
        db: EntityDatabase,
        task_queue_writer: TaskQueueWriter,
        platform_config: PlatformConfig,
        scheduler: Scheduler,
        elog: Optional[EventGroupType] = None,
    ) -> None:
        self.log = logging.getLogger("SnapShooter")

        self.db = db
        self.task_queue_writer = task_queue_writer
        self.model_spec = platform_config.model_spec
        self.entity_relation_attrs = defaultdict(dict)
        for (entity, attr), _ in self.model_spec.relations.items():
            self.entity_relation_attrs[entity][attr] = True
        for entity in self.model_spec.entities:
            self.entity_relation_attrs[entity]["_id"] = True

        self.worker_index = platform_config.process_index
        self.worker_cnt = platform_config.num_processes
        self.config = SnapShooterConfig.model_validate(platform_config.config.get("snapshots"))

        self.elog = elog or DummyEventGroup()

        self._timeseries_hooks = SnapshotTimeseriesHookContainer(self.log, self.model_spec, elog)
        self._correlation_hooks = SnapshotCorrelationHookContainer(self.log, self.model_spec, elog)
        self._init_hooks: list[Callable[[], list[DataPointTask]]] = []
        self._finalize_hooks: list[Callable[[], list[DataPointTask]]] = []

        queue = f"{platform_config.app_name}-worker-{platform_config.process_index}-snapshots"
        self.snapshot_queue_reader = TaskQueueReader(
            callback=self.process_snapshot_task,
            parse_task=Snapshot.model_validate_json,
            app_name=platform_config.app_name,
            worker_index=platform_config.process_index,
            rabbit_config=platform_config.config.get("processing_core.msg_broker", {}),
            queue=queue,
            priority_queue=False,
            parent_logger=self.log,
        )
        scheduler.register(self.snapshot_queue_reader.watchdog, second="10,40")

        self.snapshot_entities = [
            entity for entity, spec in self.model_spec.entities.items() if spec.snapshot
        ]
        self.log.info("Snapshots will be created for entities: %s", self.snapshot_entities)
        self.link_mirrors = defaultdict(dict)
        for (entity, attr), spec in self.model_spec.relations.items():
            if spec.is_relation and spec.is_mirrored:
                self.link_mirrors[spec.relation_to][spec.mirror_as] = (entity, attr)
        self.log.info("Link mirrors: %s", self.link_mirrors)

        # Get link cache
        self.cache = self.db.get_module_cache("Link")

        if platform_config.process_index != 0:
            self.log.debug(
                "Snapshot task creation will be disabled in this worker to avoid race conditions."
            )
            self.snapshot_queue_writer = None
            return

        self.snapshot_queue_writer = TaskQueueWriter(
            platform_config.app_name,
            platform_config.num_processes,
            platform_config.config.get("processing_core.msg_broker"),
            f"{platform_config.app_name}-main-snapshot-exchange",
            parent_logger=self.log,
        )

        # Schedule snapshot period
        snapshot_cron = self.config.creation_rate.model_dump(exclude_none=True)
        scheduler.register(self.make_snapshots, **snapshot_cron)

    def start(self):
        """Connect to RabbitMQ and start consuming from TaskQueue."""
        self.log.info("Connecting to RabbitMQ")
        self.snapshot_queue_reader.connect()
        self.snapshot_queue_reader.check()  # check presence of needed queues
        if self.snapshot_queue_writer is not None:
            self.snapshot_queue_writer.connect()
            self.snapshot_queue_writer.check()  # check presence of needed exchanges

        self.snapshot_queue_reader.start()
        self.log.info("Following used links detected: %s", self._correlation_hooks.used_links)
        self.log.info("SnapShooter started.")

    def stop(self):
        """Stop consuming from TaskQueue, disconnect from RabbitMQ."""
        self.snapshot_queue_reader.stop()

        if self.snapshot_queue_writer is not None:
            self.snapshot_queue_writer.disconnect()
        self.snapshot_queue_reader.disconnect()

    def register_timeseries_hook(
        self,
        hook: Callable[[str, str, list[dict]], list[DataPointTask]],
        entity_type: str,
        attr_type: str,
    ):
        """
        Registers passed timeseries hook to be called during snapshot creation.

        Binds hook to specified `entity_type` and `attr_type` (though same hook can be bound
        multiple times).

        Args:
            hook: `hook` callable should expect entity_type, attr_type and attribute
                history as arguments and return a list of `DataPointTask` objects.
            entity_type: specifies entity type
            attr_type: specifies attribute type

        Raises:
            ValueError: If entity_type and attr_type do not specify a valid timeseries attribute,
                a ValueError is raised.
        """
        self._timeseries_hooks.register(hook, entity_type, attr_type)

    def register_correlation_hook(
        self,
        hook: Callable[[str, dict], Union[None, list[DataPointTask]]],
        entity_type: str,
        depends_on: list[list[str]],
        may_change: list[list[str]],
    ):
        """
        Registers passed hook to be called during snapshot creation.

        Binds hook to specified entity_type (though same hook can be bound multiple times).

        `entity_type` and attribute specifications are validated, `ValueError` is raised on failure.

        Args:
            hook: `hook` callable should expect entity type as str
                and its current values, including linked entities, as dict
                Can optionally return a list of DataPointTask objects to perform.
            entity_type: specifies entity type
            depends_on: each item should specify an attribute that is depended on
                in the form of a path from the specified entity_type to individual attributes
                (even on linked entities).
            may_change: each item should specify an attribute that `hook` may change.
                specification format is identical to `depends_on`.

        Raises:
            ValueError: On failure of specification validation.
        """
        self._correlation_hooks.register(hook, entity_type, depends_on, may_change)

    def register_run_init_hook(self, hook: Callable[[], list[DataPointTask]]):
        """
        Registers passed hook to be called before a run of  snapshot creation begins.

        Args:
            hook: `hook` callable should expect no arguments and
                return a list of DataPointTask objects to perform.
        """
        self._init_hooks.append(hook)

    def register_run_finalize_hook(self, hook: Callable[[], list[DataPointTask]]):
        """
        Registers passed hook to be called after a run of  snapshot creation ends.

        Args:
            hook: `hook` callable should expect no arguments and
                return a list of DataPointTask objects to perform.
        """
        self._finalize_hooks.append(hook)

    def make_snapshots(self):
        """Creates snapshots for all entities currently active in database."""
        time = datetime.utcnow()
        self.db.save_metadata(
            time,
            {
                "task_creation_start": time,
                "entities": 0,
                "components": 0,
                "workers_finished": 0,
                "linked_finished": False,
            },
        )
        # Broadcast run start
        self.snapshot_queue_writer.broadcast_task(
            task=Snapshot(type=SnapshotMessageType.run_start, time=time)
        )

        # distribute list of possibly linked entities to all workers
        cached = self.get_cached_link_entity_ids()
        self.log.debug("Broadcasting %s cached linked entities", len(cached))
        self.snapshot_queue_writer.broadcast_task(
            task=Snapshot(entities=cached, time=time, type=SnapshotMessageType.linked_entities)
        )

        # Load links only for a reduced set of entities
        self.log.debug("Loading linked entities.")

        times = {}
        counts = {"entities": 0, "components": 0}
        try:
            linked_entities = self.get_linked_entities(time, cached)
            times["components_loaded"] = datetime.utcnow()

            for i, linked_entities_component in enumerate(linked_entities):
                counts["entities"] += len(linked_entities_component)
                counts["components"] += 1

                self.snapshot_queue_writer.put_task(
                    task=Snapshot(
                        entities=linked_entities_component,
                        time=time,
                        type=SnapshotMessageType.task,
                        final=(i + 1 == len(linked_entities)),
                    )
                )

            if len(linked_entities) == 0:
                self.db.update_metadata(time, metadata={"linked_finished": True})
        except pymongo.errors.CursorNotFound as err:
            self.log.exception(err)
        finally:
            times["task_creation_end"] = datetime.utcnow()
            self.db.update_metadata(
                time,
                metadata=times,
                increase=counts,
            )

        # Broadcast run end
        self.snapshot_queue_writer.broadcast_task(
            task=Snapshot(type=SnapshotMessageType.run_end, time=time)
        )

    def get_cached_link_entity_ids(self):
        used = [f"{etype}#{attr}" for etype, attr in self._correlation_hooks.used_links]
        result = self.cache.aggregate(
            [{"$match": {"using_attr": {"$in": used}}}, {"$group": {"_id": "$from"}}]
        )
        links_from = {tuple(doc["_id"].split("#", maxsplit=1)) for doc in result}
        result = self.cache.aggregate(
            [{"$match": {"using_attr": {"$in": used}}}, {"$group": {"_id": "$to"}}]
        )
        links_to = {tuple(doc["_id"].split("#", maxsplit=1)) for doc in result}
        return list(links_from | links_to)

    def get_linked_entities(self, time: datetime, cached_linked_entities: list[tuple[str, str]]):
        """Get weakly connected components from entity graph."""
        visited_entities = set()
        entity_to_component = {}
        linked_components = []
        for etype, eid in cached_linked_entities:
            master_record = self.db.get_master_record(
                etype, eid, projection=self.entity_relation_attrs[etype]
            ) or {"_id": eid}

            if not self.config.keep_empty and len(master_record) == 1:
                continue

            if (etype, master_record["_id"]) not in visited_entities:
                # Get entities linked by current entity
                current_values = self.get_values_at_time(etype, master_record, time)
                linked_entities = self.load_linked_entity_ids(etype, current_values, time)

                # Set linked as visited
                visited_entities.update(linked_entities)

                # Update component, take all connected components into account
                have_component = linked_entities & set(entity_to_component.keys())
                if have_component:
                    for entity in have_component:
                        component = entity_to_component[entity]
                        linked_entities.update(component)
                entity_to_component.update({entity: linked_entities for entity in linked_entities})

        # Make a list of unique components
        visited_entities.clear()
        for entity, component in entity_to_component.items():
            if entity in visited_entities:
                continue
            visited_entities.update(component)
            linked_components.append(component)

        return linked_components

    def process_snapshot_task(self, msg_id, task: Snapshot):
        """
        Acknowledges the received message and makes a snapshot according to the `task`.

        This function should not be called directly, but set as callback for TaskQueueReader.
        """
        if not self.snapshot_queue_reader.ack(msg_id):
            self.log.warning("Acking message %s failed, will not process.", msg_id)
            return
        if task.type == SnapshotMessageType.task:
            self.make_snapshot(task)
        elif task.type == SnapshotMessageType.linked_entities:
            self.make_snapshots_by_hash(task)
        elif task.type == SnapshotMessageType.run_start:
            self.log.debug("Run start, running init hooks")
            self._run_hooks(self._init_hooks)
        elif task.type == SnapshotMessageType.run_end:
            self.log.debug("Run end, running finalize hooks")
            self._run_hooks(self._finalize_hooks)
        else:
            raise ValueError("Unknown SnapshotMessageType.")

    def _run_hooks(self, hooks: list[Callable[[], list[DataPointTask]]]):
        tasks = []
        with task_context(self.model_spec):
            for hook in hooks:
                self.log.debug("Running hook: '%s'", get_func_name(hook))
                try:
                    new_tasks = hook()
                    tasks.extend(new_tasks)
                except Exception as e:
                    self.elog.log("module_error")
                    self.log.error(f"Error during running hook {hook}: {e}")

        for task in tasks:
            self.task_queue_writer.put_task(task)

    def make_snapshots_by_hash(self, task: Snapshot):
        """
        Make snapshots for all entities with routing key belonging to this worker.
        """
        self.log.debug("Creating snapshots for worker portion by hash.")
        have_links = set(task.entities)
        entity_cnt = 0
        for etype in self.snapshot_entities:
            records_cursor = self.db.get_worker_master_records(
                self.worker_index, self.worker_cnt, etype, no_cursor_timeout=True
            )
            for attempt in range(RETRY_COUNT):
                try:
                    entity_cnt += self.make_linkless_snapshots(
                        etype, records_cursor, task.time, have_links
                    )
                except Exception as err:
                    self.log.exception("Uncaught exception while creating snapshots: %s", err)
                    if attempt < RETRY_COUNT - 1:
                        self.log.info("Retrying snapshot creation for '%s' due to errors.", etype)
                    continue
                finally:
                    records_cursor.close()
                break
            else:
                self.log.error(
                    "Failed to create snapshots for '%s' after %s attempts.", etype, attempt + 1
                )
        self.db.update_metadata(
            task.time,
            metadata={},
            increase={"entities": entity_cnt, "components": entity_cnt, "workers_finished": 1},
            worker_id=0,
        )
        self.log.debug("Worker snapshot creation done.")

    def make_linkless_snapshots(
        self, etype: str, records_cursor, time: datetime, have_links: set
    ) -> int:
        entity_cnt = 0
        snapshots = []
        for master_record in records_cursor:
            if (etype, master_record["_id"]) in have_links:
                continue
            if not self.config.keep_empty and len(master_record) == 1:
                continue
            entity_cnt += 1
            snapshots.append(self.make_linkless_snapshot(etype, master_record, time))
            if len(snapshots) >= DB_SEND_CHUNK:
                self.db.save_snapshots(etype, snapshots, time)
                snapshots.clear()

        if snapshots:
            self.db.save_snapshots(etype, snapshots, time)
            snapshots.clear()
        return entity_cnt

    def make_linkless_snapshot(self, entity_type: str, master_record: dict, time: datetime):
        """
        Make a snapshot for given entity `master_record` and `time`.

        Runs timeseries and correlation hooks.
        The resulting snapshot is saved into DB.
        """
        self.run_timeseries_processing(entity_type, master_record)
        values = self.get_values_at_time(entity_type, master_record, time)
        self.add_mirrored_links(entity_type, values)
        entity_values = {(entity_type, master_record["_id"]): values}

        tasks = self._correlation_hooks.run(entity_values)
        for task in tasks:
            self.task_queue_writer.put_task(task)

        assert len(entity_values) == 1, "Expected a single entity."
        for record in entity_values.values():
            return record

    def add_mirrored_links(self, entity_type: str, values: dict):
        """
        This function adds mirrored links to the dict with current values of an entity.

        The links are added in the same format as normal links, i.e. as a list of dicts.
        """
        if entity_type in self.link_mirrors:
            for mirror_name, (etype, attr) in self.link_mirrors[entity_type].items():
                link_sources = self.cache.aggregate(
                    [
                        {
                            "$match": {
                                "to": f"{entity_type}#{values['eid']}",
                                "using_attr": f"{etype}#{attr}",
                            }
                        },
                        {"$project": {"from": 1}},
                    ]
                )
                values[mirror_name] = [
                    {"eid": source["from"].split("#", maxsplit=1)[1]} for source in link_sources
                ]

    def make_snapshot(self, task: Snapshot):
        """
        Make a snapshot for entities and time specified by `task`.

        Runs timeseries and correlation hooks.
        The resulting snapshots are saved into DB.
        """
        entity_values = {}
        for entity_type, entity_id in task.entities:
            record = self.db.get_master_record(entity_type, entity_id) or {"_id": entity_id}
            if not self.config.keep_empty and len(record) == 1:
                continue

            self.run_timeseries_processing(entity_type, record)
            values = self.get_values_at_time(entity_type, record, task.time)
            self.add_mirrored_links(entity_type, values)
            entity_values[entity_type, entity_id] = values

        self.link_loaded_entities(entity_values)
        created_tasks = self._correlation_hooks.run(entity_values)
        for created_task in created_tasks:
            self.task_queue_writer.put_task(created_task)

        # unlink entities again
        for (rtype, _rid), record in entity_values.items():
            for attr, value in record.items():
                if (rtype, attr) not in self.model_spec.relations:
                    continue
                spec = self.model_spec.relations[rtype, attr]
                if spec.t == AttrType.OBSERVATIONS and spec.multi_value:
                    for val in value:
                        self._remove_record_from_value(spec, val)
                else:
                    self._remove_record_from_value(spec, value)

        for (rtype, _rid), record in entity_values.items():
            if len(record) == 1 and not self.config.keep_empty:
                continue
            self.db.save_snapshot(rtype, record, task.time)

        if task.final:
            self.db.update_metadata(task.time, metadata={"linked_finished": True}, worker_id=0)

    @staticmethod
    def _remove_record_from_value(spec: AttrSpecType, value: Union[dict, list[dict]]):
        if spec.is_iterable:
            for link_dict in value:
                if "record" in link_dict:
                    del link_dict["record"]
        elif "record" in value:
            del value["record"]

    def run_timeseries_processing(self, entity_type, master_record):
        """
        - all registered timeseries processing modules must be called
          - this should result in `observations` or `plain` datapoints, which will be saved to db
            and forwarded in processing
        """
        tasks = []
        for attr, attr_spec in self.model_spec.entity_attributes[entity_type].items():
            if attr_spec.t == AttrType.TIMESERIES and attr in master_record:
                new_tasks = self._timeseries_hooks.run(entity_type, attr, master_record[attr])
                tasks.extend(new_tasks)

        self.extend_master_record(entity_type, master_record, tasks)
        for task in tasks:
            self.task_queue_writer.put_task(task)

    @staticmethod
    def extend_master_record(etype, master_record, new_tasks: list[DataPointTask]):
        """Update existing master record with datapoints from new tasks"""
        for task in new_tasks:
            for datapoint in task.data_points:
                if datapoint.etype != etype:
                    continue
                dp_dict = datapoint.dict(include={"v", "t1", "t2", "c"})
                if datapoint.attr in master_record:
                    master_record[datapoint.attr].append()
                else:
                    master_record[datapoint.attr] = [dp_dict]

    def get_values_at_time(self, etype: str, master_record: dict, time: datetime) -> dict:
        current_values = {"eid": master_record["_id"]}
        for attr, attr_spec in self.model_spec.entity_attributes[etype].items():
            if (
                attr not in master_record
                or attr_spec.t not in AttrType.PLAIN | AttrType.OBSERVATIONS
            ):
                continue
            if attr_spec.t == AttrType.PLAIN:
                current_values[attr] = master_record[attr]["v"]
                continue

            if attr_spec.multi_value:
                val, conf = self.get_multi_value_at_time(attr_spec, master_record[attr], time)
            else:
                val, conf = self.get_value_at_time(attr_spec, master_record[attr], time)

            if conf:  # conf != 0.0 or len(conf) > 0
                current_values[attr] = val
                current_values[f"{attr}#c"] = conf
        return current_values

    def load_linked_entity_ids(self, entity_type: str, current_values: dict, time: datetime):
        """
        Loads the subgraph of entities linked to the current entity,
        returns a list of their types and ids.
        """
        loaded_entity_ids = {(entity_type, current_values["eid"])}
        linked_entity_ids_to_process = (
            self.get_linked_entity_ids(entity_type, current_values) - loaded_entity_ids
        )

        while linked_entity_ids_to_process:
            entity_identifiers = linked_entity_ids_to_process.pop()
            linked_etype, linked_eid = entity_identifiers
            relevant_attributes = self.entity_relation_attrs[linked_etype]
            record = self.db.get_master_record(
                linked_etype, linked_eid, projection=relevant_attributes
            ) or {"_id": linked_eid}
            linked_values = self.get_values_at_time(linked_etype, record, time)

            linked_entity_ids_to_process.update(
                self.get_linked_entity_ids(entity_type, linked_values) - set(loaded_entity_ids)
            )
            loaded_entity_ids.add((linked_etype, linked_eid))

        return loaded_entity_ids

    def get_linked_entity_ids(self, entity_type: str, current_values: dict) -> set[tuple[str, str]]:
        """
        Returns a set of tuples (entity_type, entity_id) identifying entities linked by
        `current_values`.
        """
        related_entity_ids = set()
        for attr, val in current_values.items():
            if (entity_type, attr) not in self._correlation_hooks.used_links:
                continue
            attr_spec = self.model_spec.relations[entity_type, attr]
            if attr_spec.t == AttrType.OBSERVATIONS and attr_spec.multi_value:
                for v in val:
                    related_entity_ids.update(self._get_link_entity_ids(attr_spec, v))
            else:
                related_entity_ids.update(self._get_link_entity_ids(attr_spec, val))
        return related_entity_ids

    @staticmethod
    def _get_link_entity_ids(
        spec: AttrSpecType, link_value: Union[list[dict], dict]
    ) -> set[tuple[str, str]]:
        if spec.is_iterable:
            return {(spec.relation_to, v["eid"]) for v in link_value}
        return {(spec.relation_to, link_value["eid"])}

    def link_loaded_entities(self, loaded_entities: dict):
        for (entity_type, _entity_id), entity in loaded_entities.items():
            del_keys = []
            for attr, val in entity.items():
                if (entity_type, attr) not in self._correlation_hooks.used_links:
                    continue
                attr_spec = self.model_spec.relations[entity_type, attr]
                if attr_spec.t == AttrType.OBSERVATIONS and attr_spec.multi_value:
                    entity[attr] = []
                    val_conf = entity[f"{attr}#c"]
                    pruned_conf = []
                    for v, conf in zip(val, val_conf):
                        if self._keep_link(loaded_entities, attr_spec, v):
                            self._link_record(loaded_entities, attr_spec, v)
                            entity[attr].append(v)
                            pruned_conf.append(conf)
                    entity[f"{attr}#c"] = pruned_conf
                elif self._keep_link(loaded_entities, attr_spec, val):
                    self._link_record(loaded_entities, attr_spec, val)
                    entity[attr] = val
                else:  # The linked entity does not exist, and we do not want to keep empty
                    del_keys.append(attr)
                    if f"{attr}#c" in entity:
                        del_keys.append(f"{attr}#c")

            for key in del_keys:
                del entity[key]

    def _keep_link(
        self, loaded_entities: dict, attr_spec: AttrSpecType, val: Union[dict, list[dict]]
    ) -> bool:
        if self.config.keep_empty:
            return True
        if attr_spec.is_iterable:
            return any(
                loaded_entities.get((attr_spec.relation_to, v["eid"])) is not None for v in val
            )
        return loaded_entities.get((attr_spec.relation_to, val["eid"])) is not None

    @staticmethod
    def _link_record(loaded_entities: dict, attr_spec: AttrSpecType, val: Union[dict, list[dict]]):
        if attr_spec.is_iterable:
            for link_dict in val:
                link_dict["record"] = loaded_entities.get(
                    (attr_spec.relation_to, link_dict["eid"]), {"eid": link_dict["eid"]}
                )
        else:
            val["record"] = loaded_entities.get(
                (attr_spec.relation_to, val["eid"]), {"eid": val["eid"]}
            )

    def get_value_at_time(
        self, attr_spec: AttrSpecObservations, attr_history, time: datetime
    ) -> tuple[Any, float]:
        """Get current value of an attribute from its history. Assumes `multi_value = False`."""
        return max(
            (
                (point["v"], self.extrapolate_confidence(point, time, attr_spec.history_params))
                for point in attr_history
            ),
            key=lambda val_conf: val_conf[1],
            default=(None, 0.0),
        )

    def get_multi_value_at_time(
        self, attr_spec: AttrSpecObservations, attr_history, time: datetime
    ) -> tuple[list, list[float]]:
        """Get current value of a multi_value attribute from its history."""
        if attr_spec.data_type.hashable:
            values_with_confidence = defaultdict(float)
            for point in attr_history:
                value = point["v"]
                confidence = self.extrapolate_confidence(point, time, attr_spec.history_params)
                if confidence > 0.0 and values_with_confidence[value] < confidence:
                    values_with_confidence[value] = confidence
            return list(values_with_confidence.keys()), list(values_with_confidence.values())
        else:
            values = []
            confidence_list = []
            for point in attr_history:
                value = point["v"]
                confidence = self.extrapolate_confidence(point, time, attr_spec.history_params)
                if value in values:
                    i = values.index(value)
                    confidence_list[i] = max(confidence_list[i], confidence)
                elif confidence > 0.0:
                    values.append(value)
                    confidence_list.append(confidence)
            return values, confidence_list

    @staticmethod
    def extrapolate_confidence(
        datapoint: dict, time: datetime, history_params: ObservationsHistoryParams
    ) -> float:
        """Get the confidence value at given time."""
        t1 = datapoint["t1"]
        t2 = datapoint["t2"]
        base_confidence = datapoint["c"]

        if time < t1:
            if time <= t1 - history_params.pre_validity:
                return 0.0
            return base_confidence * (1 - (t1 - time) / history_params.pre_validity)
        if time <= t2:
            return base_confidence  # completely inside the (strict) interval
        if time >= t2 + history_params.post_validity:
            return 0.0
        return base_confidence * (1 - (time - t2) / history_params.post_validity)



# Very simple datapoint generator for bus example config

import random
from argparse import ArgumentParser
from datetime import datetime, timedelta
from sys import stderr
from time import sleep

import requests


def random_initial_location():
    latitude = random.uniform(39.0, 41.0)
    longitute = random.uniform(18.0, 22.0)

    return [round(latitude, 3), round(longitute, 3)]


def do_random_location_increment(current_location):
    current_location[0] = round(current_location[0] + random.uniform(-1.0, 1.0), 3)
    current_location[1] = round(current_location[1] + random.uniform(-1.0, 1.0), 3)


def t1():
    return datetime.utcnow()


def random_t2():
    return datetime.utcnow() + timedelta(minutes=random.randint(5, 15))


def random_passenger_counts_3():
    return [random.randint(0, 40), random.randint(0, 40), random.randint(0, 40)]


bus_lines = {
    "10": {"label": "Special bus 10", "location": random_initial_location()},
    "11": {"label": "Special bus 11", "location": random_initial_location()},
    "12": {"label": "Bus 12", "location": random_initial_location()},
    "19": {"label": "Bus 19 ", "location": random_initial_location()},
    "48": {"label": "Bus 48", "location": random_initial_location()},
    "93": {"label": "Bus 93", "location": random_initial_location()},
    "E73": {"label": "Express bus 73", "location": random_initial_location()},
    "N72": {"label": "Night bus 72", "location": random_initial_location()},
    "S20": {"label": "School bus 20", "location": random_initial_location()},
}


if __name__ == "__main__":
    # Parse arguments
    parser = ArgumentParser("Simple datapoint generator for DP3 testing.")
    parser.add_argument(
        "--endpoint-url",
        help="Base DP3 API endpoint. Default='http://127.0.0.1:5000'.",
        default="http://127.0.0.1:5000",
        type=str,
    )
    parser.add_argument(
        "--interval",
        help="Interval of sending datapoints in minutes. Default: 1 minute.",
        default=1,
        type=int,
    )
    args = parser.parse_args()

    # API endpoint
    datapoints_url = f"{args.endpoint_url}/datapoints"

    # Datapoints buffer
    dps = []

    # One-time datapoints
    for n in bus_lines:
        bus_line = bus_lines[n]

        dps.append(
            {
                "type": "bus",
                "id": n,
                "attr": "label",
                "v": bus_line["label"],
                "t1": t1().isoformat(),
                "t2": random_t2().isoformat(),
                "src": "Manual",
            }
        )

    # Realtime datapoints
    while True:
        for n in bus_lines:
            bus_line = bus_lines[n]

            dps.append(
                {
                    "type": "bus",
                    "id": n,
                    "attr": "location",
                    "v": bus_line["location"],
                    "t1": t1().isoformat(),
                    "t2": random_t2().isoformat(),
                    "src": "Tracking SW",
                }
            )
            do_random_location_increment(bus_line["location"])

            dps.append(
                {
                    "type": "bus",
                    "id": n,
                    "attr": "speed",
                    "v": round(random.uniform(0.0, 100.0), 1),
                    "t1": t1().isoformat(),
                    "t2": random_t2().isoformat(),
                    "src": "Tracking SW",
                }
            )

            t1_local = t1()
            dps.append(
                {
                    "type": "bus",
                    "id": n,
                    "attr": "passengers_in_out",
                    "v": {
                        "front_in": random_passenger_counts_3(),
                        "front_out": random_passenger_counts_3(),
                        "middle_in": random_passenger_counts_3(),
                        "middle_out": random_passenger_counts_3(),
                        "back_in": random_passenger_counts_3(),
                        "back_out": random_passenger_counts_3(),
                    },
                    "t1": t1_local.isoformat(),
                    "t2": (t1_local + timedelta(minutes=30)).isoformat(),
                    "src": "Bus counter",
                }
            )

        # Send datapoints
        try:
            response = requests.post(url=datapoints_url, json=dps)

            if response.status_code == requests.codes.ok:
                print(f"Sent {len(dps)} datapoints of attribute(s)")
            else:
                response_content = response.content.decode("utf-8")
                print(
                    f"Sending datapoints failed: {response.reason} {response_content}",
                    file=stderr,
                )
                print(f"Payload: {dps}", file=stderr)

        except requests.exceptions.ConnectionError:
            print("Sending datapoints failed: connection error", file=stderr)

        except KeyboardInterrupt:
            print("Received keyboard interrupt, exiting.", file=stderr)
            break

        # Clear buffer
        dps = []

        # Wait for next cycle
        sleep(60 * args.interval)



import common

from dp3.api.internal.entity_response_models import EntityEidAttrValue, EntityEidAttrValueOrHistory

TESTED_PATH = "entity/test_entity_type/eid01/{action}/test_attr_string"


class SetEidAttrValue(common.APITest):
    def test_invalid_payload(self):
        response = self.post_request(TESTED_PATH.format(action="set"), json={})
        self.assertEqual(response.status_code, 422, msg=response.json())

    def test_valid_payload(self):
        payload = EntityEidAttrValue(value="Test string val1")
        response = self.post_request(
            TESTED_PATH.format(action="set"), data=payload.model_dump_json()
        )
        self.assertEqual(response.status_code, 200, msg=response.json())

        expected = EntityEidAttrValueOrHistory(attr_type=1, current_value="Test string val1")
        self.query_expected_value(
            lambda: self.get_entity_data(
                TESTED_PATH.format(action="get"), EntityEidAttrValueOrHistory
            ),
            lambda received: received == expected,
        )



from fastapi import HTTPException
from pydantic import BaseModel


class HealthCheckResponse(BaseModel):
    """Healthcheck endpoint response"""

    detail: str = "It works!"


class SuccessResponse(BaseModel):
    """Generic success response"""

    detail: str = "OK"


class ErrorResponse(BaseModel):
    """Generic error response"""

    detail: str


class RequestValidationError(HTTPException):
    """HTTP exception wrapper to simplify path and query validation"""

    def __init__(self, loc, msg):
        super().__init__(422, [{"loc": loc, "msg": msg, "type": "value_error"}])



#!/usr/bin/env python3
"""Code of the main worker process.

Don't run directly. Import and run the main() function.
"""
import inspect
import logging
import os
import signal
import sys
import threading
from functools import partial
from importlib import import_module

from event_count_logger import DummyEventGroup, EventCountLogger
from pydantic import ValidationError

from dp3.common.callback_registrar import CallbackRegistrar, reload_module_config
from dp3.common.config import PlatformConfig
from dp3.common.control import Control, ControlAction, refresh_on_entity_creation
from dp3.core.collector import GarbageCollector
from dp3.core.link_manager import LinkManager
from dp3.core.updater import Updater
from dp3.history_management.telemetry import Telemetry
from dp3.task_processing.task_queue import TaskQueueWriter

if __name__ == "__main__":
    print("Don't run this file directly. Use 'bin/worker' instead.", file=sys.stderr)
    sys.exit(1)

from .common import scheduler
from .common.base_module import BaseModule
from .common.config import ModelSpec, read_config_dir
from .database.database import EntityDatabase
from .history_management.history_manager import HistoryManager
from .snapshots.snapshooter import SnapShooter
from .task_processing.task_distributor import TaskDistributor
from .task_processing.task_executor import TaskExecutor


def load_modules(
    modules_dir: str,
    enabled_modules: str,
    log: logging.Logger,
    registrar: CallbackRegistrar,
    platform_config: PlatformConfig,
) -> dict[str, BaseModule]:
    """Load plug-in modules

    Import Python modules with names in 'enabled_modules' from 'modules_dir' directory
    and return all found classes derived from BaseModule class.
    """
    # Get list of all modules available in given folder
    # [:-3] is for removing '.py' suffix from module filenames
    available_modules = []
    for item in os.scandir(modules_dir):
        # A module can be a Python file or a Python package
        # (i.e. a directory with "__init__.py" file)
        if item.is_file() and item.name.endswith(".py"):
            available_modules.append(item.name[:-3])  # name without .py
        if item.is_dir() and "__init__.py" in os.listdir(os.path.join(modules_dir, item.name)):
            available_modules.append(item.name)

    log.debug(f"Available modules: {', '.join(available_modules)}")
    log.debug(f"Enabled modules: {', '.join(enabled_modules)}")

    # Check if all desired modules are in modules folder
    missing_modules = set(enabled_modules) - set(available_modules)
    if missing_modules:
        log.fatal(
            "Some of desired modules are not available (not in modules folder), "
            f"specifically: {missing_modules}"
        )
        sys.exit(2)

    # Do imports of desired modules from 'modules' folder
    # (rewrite sys.path to modules_dir, import all modules and rewrite it back)
    log.debug("Importing modules ...")
    sys.path.insert(0, modules_dir)
    imported_modules: list[tuple[str, str, type[BaseModule]]] = [
        (module_name, name, obj)
        for module_name in enabled_modules
        for name, obj in inspect.getmembers(import_module(module_name))
        if inspect.isclass(obj) and BaseModule in obj.__bases__
    ]
    del sys.path[0]

    # Loaded modules dict will contain main classes from all desired modules,
    # which has BaseModule as parent
    modules_main_objects = {}
    for module_name, _, obj in imported_modules:
        # Append instance of module class (obj is class --> obj() is instance)
        # --> call init, which registers handler
        module_config = platform_config.config.get(f"modules.{module_name}", {})
        modules_main_objects[module_name] = obj(platform_config, module_config, registrar)
        log.info(f"Module loaded: {module_name}:{obj.__name__}")

    return modules_main_objects


def main(app_name: str, config_dir: str, process_index: int, verbose: bool) -> None:
    """
    Run worker process.
    Args:
        app_name: Name of the application to distinct it from other DP3-based apps.
            For example, it's used as a prefix for RabbitMQ queue names.
        config_dir: Path to directory containing configuration files.
        process_index: Index of this worker process. For each application
            there must be N processes running simultaneously, each started with a
            unique index (from 0 to N-1). N is read from configuration
            ('worker_processes' in 'processing_core.yml').
        verbose: More verbose output (set log level to DEBUG).
    """
    ##############################################
    # Initialize logging mechanism
    threading.current_thread().name = f"MainThread-{process_index}"
    LOGFORMAT = "%(asctime)-15s,%(threadName)s,%(name)s,[%(levelname)s] %(message)s"
    LOGDATEFORMAT = "%Y-%m-%dT%H:%M:%S"

    logging.basicConfig(
        level=logging.DEBUG if verbose else logging.INFO, format=LOGFORMAT, datefmt=LOGDATEFORMAT
    )
    log = logging.getLogger()

    # Disable INFO and DEBUG messages from some libraries
    logging.getLogger("requests").setLevel(logging.WARNING)
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("amqpstorm").setLevel(logging.WARNING)

    ##############################################
    # Load configuration
    config_base_path = os.path.abspath(config_dir)
    log.debug(f"Loading config directory {config_base_path}")

    # Whole configuration should be loaded
    config = read_config_dir(config_base_path, recursive=True)
    try:
        model_spec = ModelSpec(config.get("db_entities"))
    except ValidationError as e:
        log.fatal("Invalid model specification: %s", e)
        sys.exit(2)

    # Print whole attribute specification
    log.debug(model_spec)

    num_processes = config.get("processing_core.worker_processes")

    platform_config = PlatformConfig(
        app_name=app_name,
        config_base_path=config_base_path,
        config=config,
        model_spec=model_spec,
        process_index=process_index,
        num_processes=num_processes,
    )
    ##############################################
    # Create instances of core components
    log.info(f"***** {app_name} worker {process_index} of {num_processes} start *****")

    # EventCountLogger
    ecl = EventCountLogger(
        platform_config.config.get("event_logging.groups"),
        platform_config.config.get("event_logging.redis"),
    )
    elog = ecl.get_group("te") or DummyEventGroup()
    elog_by_src = ecl.get_group("tasks_by_src") or DummyEventGroup()

    db = EntityDatabase(config, model_spec, num_processes, process_index, elog)
    if process_index == 0:
        db.update_schema()
    else:
        db.await_updated_schema()

    global_scheduler = scheduler.Scheduler()
    task_executor = TaskExecutor(db, platform_config, elog, elog_by_src)
    snap_shooter = SnapShooter(
        db,
        TaskQueueWriter(app_name, num_processes, config.get("processing_core.msg_broker")),
        platform_config,
        global_scheduler,
        elog,
    )
    updater = Updater(
        db,
        TaskQueueWriter(app_name, num_processes, config.get("processing_core.msg_broker")),
        platform_config,
        global_scheduler,
        elog,
    )
    registrar = CallbackRegistrar(global_scheduler, task_executor, snap_shooter, updater)

    LinkManager(db, platform_config, registrar)
    HistoryManager(db, platform_config, registrar)
    Telemetry(db, platform_config, registrar)
    GarbageCollector(db, platform_config, registrar)

    # Lock used to control when the program stops.
    daemon_stop_lock = threading.Lock()
    daemon_stop_lock.acquire()

    # Signal handler releasing the lock on SIGINT or SIGTERM
    def sigint_handler(signum, frame):
        log.debug(
            "Signal {} received, stopping worker".format(
                {signal.SIGINT: "SIGINT", signal.SIGTERM: "SIGTERM"}.get(signum, signum)
            )
        )
        daemon_stop_lock.release()

    signal.signal(signal.SIGINT, sigint_handler)
    signal.signal(signal.SIGTERM, sigint_handler)
    signal.signal(signal.SIGABRT, sigint_handler)

    task_distributor = TaskDistributor(task_executor, platform_config, registrar, daemon_stop_lock)

    control = Control(platform_config)
    control.set_action_handler(ControlAction.make_snapshots, snap_shooter.make_snapshots)
    control.set_action_handler(
        ControlAction.refresh_on_entity_creation,
        partial(refresh_on_entity_creation, task_distributor, task_executor),
    )
    modules = {}
    control.set_action_handler(
        ControlAction.refresh_module_config,
        partial(reload_module_config, log, platform_config, modules),
    )
    global_scheduler.register(control.control_queue.watchdog, second="15,45")

    ##############################################
    # Load all plug-in modules

    module_dir = config.get("processing_core.modules_dir")
    module_dir = os.path.abspath(os.path.join(config_base_path, module_dir))

    loaded_modules = load_modules(
        module_dir,
        config.get("processing_core.enabled_modules"),
        log,
        registrar,
        platform_config,
    )
    modules.update(loaded_modules)

    ################################################
    # Initialization completed, run ...

    # Run update manager thread
    log.info("***** Initialization completed, starting all modules *****")

    # Run modules that have their own threads (TODO: there are no such modules, should be kept?)
    # (if they don't, the start() should do nothing)
    for module in loaded_modules.values():
        module.start()

    core_modules = [
        updater,  # Updater will throw exceptions when misconfigured (best start first)
        task_distributor,  # TaskDistributor (which starts TaskExecutors in several worker threads)
        db,
        snap_shooter,
        control,
        global_scheduler,
    ]
    running_core_modules = []

    try:
        for module in core_modules:
            module.start()
            running_core_modules.append(module)

        # Wait until someone wants to stop the program by releasing this Lock.
        # It may be a user by pressing Ctrl-C or some program module.
        # (try to acquire the lock again,
        # effectively waiting until it's released by signal handler or another thread)
        if os.name == "nt":
            # This is needed on Windows in order to catch Ctrl-C, which doesn't break the waiting.
            while not daemon_stop_lock.acquire(timeout=1):
                pass
        else:
            daemon_stop_lock.acquire()
    except Exception as e:
        log.exception(e)

    ################################################
    # Finalization & cleanup
    # Set signal handlers back to their defaults,
    # so the second Ctrl-C closes the program immediately
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    signal.signal(signal.SIGTERM, signal.SIG_DFL)
    signal.signal(signal.SIGABRT, signal.SIG_DFL)

    log.info("Stopping running components ...")
    for module in reversed(running_core_modules):
        module.stop()

    for module in loaded_modules.values():
        module.stop()

    log.info("***** Finished, main thread exiting. *****")
    logging.shutdown()



"""
Core module performing deletion of entities based on specified policy.
"""

import logging
from collections import defaultdict
from datetime import datetime, timedelta
from functools import partial

from pydantic import BaseModel

from dp3.common.attrspec import AttrType
from dp3.common.callback_registrar import CallbackRegistrar
from dp3.common.config import CronExpression, PlatformConfig
from dp3.common.datapoint import DataPointBase, DataPointObservationsBase, DataPointTimeseriesBase
from dp3.common.task import DataPointTask
from dp3.database.database import EntityDatabase

DB_SEND_CHUNK = 1000


class GarbageCollectorConfig(BaseModel):
    """The configuration of the Collector module.

    Attributes:
        collection_rate: The rate at which the collector module runs. Default is 3:00 AM every day.
    """

    collection_rate: CronExpression = CronExpression(hour="3", minute="0", second="0")


class GarbageCollector:
    """Collector module manages the lifetimes of entities based on specified policy."""

    def __init__(
        self,
        db: EntityDatabase,
        platform_config: PlatformConfig,
        registrar: CallbackRegistrar,
    ):
        self.log = logging.getLogger("GarbageCollector")
        self.model_spec = platform_config.model_spec
        self.config = GarbageCollectorConfig.model_validate(
            platform_config.config.get("garbage_collector", {})
        )
        self.db = db

        self.worker_index = platform_config.process_index
        self.num_workers = platform_config.num_processes

        # Get link cache
        self.cache = self.db.get_module_cache("Link")
        self.inverse_relations = self._get_inverse_relations()

        for entity, entity_config in self.model_spec.entities.items():
            lifetime = entity_config.lifetime
            if lifetime.type == "immortal":
                pass
            elif lifetime.type == "ttl":
                registrar.register_entity_hook(
                    "on_entity_creation",
                    partial(self.extend_ttl_on_create, base_ttl=lifetime.on_create),
                    entity,
                )

                self._register_ttl_extensions(entity, registrar, lifetime.mirror_data)

                registrar.scheduler_register(
                    self.collect_ttl,
                    func_args=[entity],
                    **self.config.collection_rate.model_dump(),
                )
            elif lifetime.type == "weak":
                if self.inverse_relations[entity]:
                    registrar.scheduler_register(
                        self.collect_weak,
                        func_args=[entity],
                        **self.config.collection_rate.model_dump(),
                    )
                else:
                    raise ValueError(
                        f"Entity {entity} has weak lifetime "
                        f"but is not referenced by any other entities."
                    )

            else:
                raise ValueError(f"Unknown lifetime type: {lifetime.type}")

    def _get_inverse_relations(self) -> dict[str, set[tuple[str, str]]]:
        inverse_relations = defaultdict(set)
        for entity_attr, attr_spec in self.model_spec.relations.items():
            inverse_relations[attr_spec.relation_to].add(entity_attr)
        return inverse_relations

    def extend_ttl_on_create(
        self, eid: str, task: DataPointTask, base_ttl: timedelta
    ) -> list[DataPointTask]:
        """Extends the TTL of the entity by the specified timedelta."""
        task = DataPointTask(
            etype=task.etype,
            eid=eid,
            ttl_tokens={"base": datetime.utcnow() + base_ttl},
        )
        return [task]

    def _register_ttl_extensions(
        self, entity: str, registrar: CallbackRegistrar, mirror_data: bool
    ):
        for attr, attr_spec in self.model_spec.entity_attributes[entity].items():
            if attr_spec.t == AttrType.TIMESERIES:
                if mirror_data and attr_spec.timeseries_params.max_age is not None:
                    if attr_spec.ttl:
                        ttl = max(attr_spec.ttl, attr_spec.timeseries_params.max_age)
                    else:
                        ttl = attr_spec.timeseries_params.max_age
                else:
                    ttl = attr_spec.ttl
                if not ttl:
                    continue

                registrar.register_attr_hook(
                    "on_new_ts_chunk",
                    partial(self.extend_timeseries_ttl, extend_by=ttl),
                    entity,
                    attr,
                )
            elif attr_spec.t == AttrType.OBSERVATIONS:
                if mirror_data and attr_spec.history_params.max_age is not None:
                    if attr_spec.ttl:
                        ttl = max(attr_spec.ttl, attr_spec.history_params.max_age)
                    else:
                        ttl = attr_spec.history_params.max_age
                else:
                    ttl = attr_spec.ttl
                if not ttl:
                    continue

                registrar.register_attr_hook(
                    "on_new_observation",
                    partial(self.extend_observations_ttl, extend_by=ttl),
                    entity,
                    attr,
                )
            elif attr_spec.t == AttrType.PLAIN:
                if not attr_spec.ttl:
                    continue

                registrar.register_attr_hook(
                    "on_new_plain",
                    partial(self.extend_plain_ttl, extend_by=attr_spec.ttl),
                    entity,
                    attr,
                )
            else:
                raise ValueError(f"Unknown attribute type: {attr_spec.t}")

    def collect_weak(self, etype: str):
        """Deletes weak entities when their last reference has expired."""
        self.log.debug("Starting removal of '%s' weak entities", etype)
        start = datetime.now()
        entities = 0
        deleted = 0

        self.db.save_metadata(
            start,
            {"entities": 0, "deleted": 0, "weak_collect_start": start, "entity": etype},
        )

        # Aggregate the cache entities by their "to" field, which contains the entity
        aggregated = self.cache.aggregate(
            [
                {"$match": {"to": {"$regex": f"^{etype}#"}}},
                {"$group": {"_id": "$to"}},
            ]
        )
        have_references = [doc["_id"].split("#", maxsplit=1)[1] for doc in aggregated]
        entities += len(have_references)

        to_delete = []
        records_cursor = self.db.get_worker_master_records(
            self.worker_index,
            self.num_workers,
            etype,
            query_filter={"_id": {"$nin": have_references}},
        )
        try:
            for master_document in records_cursor:
                entities += 1
                deleted += 1
                to_delete.append(master_document["_id"])

                if len(to_delete) >= DB_SEND_CHUNK:
                    self.db.delete_eids(etype, to_delete)
                    to_delete.clear()

            if to_delete:
                self.db.delete_eids(etype, to_delete)
                to_delete.clear()

        finally:
            records_cursor.close()

        self.db.update_metadata(
            start,
            metadata={"weak_collect_end": datetime.now()},
            increase={"entities": entities, "deleted": deleted},
        )
        self.log.info(
            "Removal of '%s' weak entities done - %s tracked, %s processed & deleted",
            etype,
            entities,
            deleted,
        )

    def collect_ttl(self, etype: str):
        """Deletes entities after their TTL lifetime has expired."""
        self.log.debug("Starting removal of '%s' entities by TTL", etype)
        start = datetime.now()
        utc_now = datetime.utcnow()
        entities = 0
        deleted = 0

        to_delete = []
        expired_ttls = {}

        self.db.save_metadata(
            start, {"entities": 0, "deleted": 0, "ttl_collect_start": start, "entity": etype}
        )

        records_cursor = self.db.get_worker_master_records(
            self.worker_index, self.num_workers, etype, no_cursor_timeout=True
        )
        try:
            for master_document in records_cursor:
                entities += 1
                if "#ttl" not in master_document:
                    continue  # TTL not set, ignore for now

                if all(ttl < utc_now for ttl in master_document["#ttl"].values()):
                    deleted += 1
                    to_delete.append(master_document["_id"])
                else:
                    eid_expired_ttls = [
                        name for name, ttl in master_document["#ttl"].items() if ttl < start
                    ]
                    if eid_expired_ttls:
                        expired_ttls[master_document["_id"]] = eid_expired_ttls

                if len(to_delete) >= DB_SEND_CHUNK:
                    self.db.delete_eids(etype, to_delete)
                    to_delete.clear()
                if len(expired_ttls) >= DB_SEND_CHUNK:
                    self.db.remove_expired_ttls(etype, expired_ttls)
                    expired_ttls.clear()

            if to_delete:
                self.db.delete_eids(etype, to_delete)
                to_delete.clear()
            if expired_ttls:
                self.db.remove_expired_ttls(etype, expired_ttls)
                expired_ttls.clear()

        finally:
            records_cursor.close()

        self.db.update_metadata(
            start,
            metadata={"ttl_collect_end": datetime.now()},
            increase={"entities": entities, "deleted": deleted},
        )
        self.log.info(
            "Removal of '%s' entities by TTL done - %s processed, %s deleted",
            etype,
            entities,
            deleted,
        )

    def extend_plain_ttl(
        self, eid: str, dp: DataPointBase, extend_by: timedelta
    ) -> list[DataPointTask]:
        """Extends the TTL of the entity by the specified timedelta."""
        now = datetime.utcnow()
        task = DataPointTask(
            etype=dp.etype,
            eid=eid,
            ttl_tokens={"data": now + extend_by},
        )
        return [task]

    def extend_observations_ttl(
        self, eid: str, dp: DataPointObservationsBase, extend_by: timedelta
    ) -> list[DataPointTask]:
        """Extends the TTL of the entity by the specified timedelta."""
        task = DataPointTask(
            etype=dp.etype,
            eid=eid,
            ttl_tokens={"data": dp.t2 + extend_by},
        )
        return [task]

    def extend_timeseries_ttl(
        self, eid: str, dp: DataPointTimeseriesBase, extend_by: timedelta
    ) -> list[DataPointTask]:
        """Extends the TTL of the entity by the specified timedelta."""
        task = DataPointTask(
            etype=dp.etype,
            eid=eid,
            ttl_tokens={"data": dp.t2 + extend_by},
        )
        return [task]



"""
Module managing registered hooks and their dependencies on one another.
"""

import logging
from collections import defaultdict
from collections.abc import Hashable
from dataclasses import dataclass, field
from typing import Callable, Union

from dp3.common.attrspec import AttrType
from dp3.common.config import ModelSpec
from dp3.common.task import DataPointTask, task_context
from dp3.common.utils import get_func_name
from dp3.task_processing.task_hooks import EventGroupType


class SnapshotTimeseriesHookContainer:
    """Container for timeseries analysis hooks"""

    def __init__(self, log: logging.Logger, model_spec: ModelSpec, elog: EventGroupType):
        self.log = log.getChild("TimeseriesHooks")
        self.elog = elog
        self.model_spec = model_spec

        self._hooks = defaultdict(list)

    def register(
        self,
        hook: Callable[[str, str, list[dict]], list[DataPointTask]],
        entity_type: str,
        attr_type: str,
    ):
        """
        Registers passed timeseries hook to be called during snapshot creation.

        Binds hook to specified entity_type and attr_type (though same hook can be bound
        multiple times).
        If entity_type and attr_type do not specify a valid timeseries attribute,
        a ValueError is raised.
        Args:
            hook: `hook` callable should expect entity_type, attr_type and attribute
                history as arguments and return a list of `Task` objects.
            entity_type: specifies entity type
            attr_type: specifies attribute type
        """
        if (entity_type, attr_type) not in self.model_spec.attributes:
            raise ValueError(f"Attribute '{attr_type}' of entity '{entity_type}' does not exist.")
        spec = self.model_spec.attributes[entity_type, attr_type]
        if spec.t != AttrType.TIMESERIES:
            raise ValueError(f"'{entity_type}.{attr_type}' is not a timeseries, but '{spec.t}'")
        self._hooks[entity_type, attr_type].append(hook)
        self.log.debug(f"Added hook: '{get_func_name(hook)}'")

    def run(
        self, entity_type: str, attr_type: str, attr_history: list[dict]
    ) -> list[DataPointTask]:
        """Runs registered hooks."""
        tasks = []
        with task_context(self.model_spec):
            for hook in self._hooks[entity_type, attr_type]:
                try:
                    new_tasks = hook(entity_type, attr_type, attr_history)
                    tasks.extend(new_tasks)
                except Exception as e:
                    self.elog.log("module_error")
                    self.log.error(f"Error during running hook {hook}: {e}")
        return tasks


class SnapshotCorrelationHookContainer:
    """Container for data fusion and correlation hooks."""

    def __init__(self, log: logging.Logger, model_spec: ModelSpec, elog: EventGroupType):
        self.log = log.getChild("CorrelationHooks")
        self.elog = elog
        self.model_spec = model_spec

        self._hooks: defaultdict[str, list[tuple[str, Callable]]] = defaultdict(list)
        self._short_hook_ids: dict = {}

        self._dependency_graph = DependencyGraph(self.log)
        self.used_links = set()

    def register(
        self,
        hook: Callable[[str, dict], Union[None, list[DataPointTask]]],
        entity_type: str,
        depends_on: list[list[str]],
        may_change: list[list[str]],
    ) -> str:
        """
        Registers passed hook to be called during snapshot creation.

        Binds hook to specified entity_type (though same hook can be bound multiple times).

        If entity_type and attribute specifications are validated
        and ValueError is raised on failure.
        Args:
            hook: `hook` callable should expect entity type as str
                and its current values, including linked entities, as dict.
                Can optionally return a list of DataPointTask objects to perform.
            entity_type: specifies entity type
            depends_on: each item should specify an attribute that is depended on
                in the form of a path from the specified entity_type to individual attributes
                (even on linked entities).
            may_change: each item should specify an attribute that `hook` may change.
                specification format is identical to `depends_on`.
        Returns:
            Generated hook id.
        """

        if entity_type not in self.model_spec.entities:
            raise ValueError(f"Entity '{entity_type}' does not exist.")

        self.used_links |= self._validate_attr_paths(entity_type, depends_on)
        self.used_links |= self._validate_attr_paths(entity_type, may_change)

        depends_on = self._get_attr_path_destinations(entity_type, depends_on)
        may_change = self._get_attr_path_destinations(entity_type, may_change)

        hook_args = f"({entity_type}, [{','.join(depends_on)}], [{','.join(may_change)}])"
        hook_id = f"{get_func_name(hook)}{hook_args}"
        self._short_hook_ids[hook_id] = hook_args
        self._dependency_graph.add_hook_dependency(hook_id, depends_on, may_change)

        self._hooks[entity_type].append((hook_id, hook))
        self._restore_hook_order(self._hooks[entity_type])

        self.log.info(f"Added hook: '{hook_id}'")
        return hook_id

    def _validate_attr_paths(
        self, base_entity: str, paths: list[list[str]]
    ) -> set[tuple[str, str]]:
        """
        Validates paths of links between entity attributes

        Returns: The set of used links.
        """
        entity_attributes = self.model_spec.entity_attributes
        used_links = set()
        for path in paths:
            curr_entity = base_entity
            position = entity_attributes[base_entity]
            for i, step in enumerate(path):
                if step not in position:
                    raise ValueError(
                        f"Invalid path '{'->'.join([base_entity] + path)}', failed on '{step}'"
                    )
                position = position[step]
                if position.is_relation:
                    if i != len(path) - 1:
                        used_links.add((curr_entity, step))
                    curr_entity = position.relation_to
                    position = entity_attributes[position.relation_to]
        return used_links

    def _get_attr_path_destinations(self, base_entity: str, paths: list[list[str]]) -> list[str]:
        """
        Normalize paths to contain only destination attributes.
        Returns:
            List of destination attributes as tuples (entity, attr).
        """
        destinations = []
        for path in paths:
            resolved_path = self._resolve_entities_in_path(base_entity, path)
            dest_entity, dest_attr = resolved_path[-1]
            destinations.append(f"{dest_entity}.{dest_attr}")
        return destinations

    def _resolve_entities_in_path(self, base_entity: str, path: list[str]) -> list[tuple[str, str]]:
        """
        Resolves attributes in the path to tuples of  (entity_name, attr_name).

        Args:
            base_entity: Entity to which the first attribute belongs.
            path: List of attribute names, all must be link type except for the last one.

        Returns:
            A resolved link path of tuples (entity_name, attr_name).
        """
        entity_attributes = self.model_spec.entity_attributes
        position = entity_attributes[base_entity]
        resolved_path = []
        for step in path:
            resolved_path.append((base_entity, step))
            position = position[step]
            if position.is_relation:
                base_entity = position.relation_to
                position = entity_attributes[position.relation_to]
        return resolved_path

    def run(self, entities: dict) -> list[DataPointTask]:
        """Runs registered hooks."""
        entity_types = {etype for etype, _ in entities}
        hook_subset = [
            (hook_id, hook, etype) for etype in entity_types for hook_id, hook in self._hooks[etype]
        ]
        topological_order = self._dependency_graph.topological_order
        hook_subset.sort(key=lambda x: topological_order.index(x[0]))
        entities_by_etype = defaultdict(dict)
        for (etype, eid), values in entities.items():
            entities_by_etype[etype][eid] = values

        created_tasks = []

        with task_context(self.model_spec):
            for hook_id, hook, etype in hook_subset:
                short_id = hook_id if len(hook_id) < 160 else self._short_hook_ids[hook_id]
                for eid, entity_values in entities_by_etype[etype].items():
                    self.log.debug("Running hook %s on entity %s", short_id, eid)
                    try:
                        tasks = hook(etype, entity_values)
                        if tasks is not None and tasks:
                            created_tasks.extend(tasks)
                    except Exception as e:
                        self.elog.log("module_error")
                        self.log.error(f"Error during running hook {hook_id}: {e}")
                        self.log.exception(e)

        return created_tasks

    def _restore_hook_order(self, hooks: list[tuple[str, Callable]]):
        topological_order = self._dependency_graph.topological_sort()
        hooks.sort(key=lambda x: topological_order.index(x[0]))


@dataclass
class GraphVertex:
    """Vertex in a graph of dependencies"""

    adj: list = field(default_factory=list)
    in_degree: int = 0
    type: str = "attr"


class DependencyGraph:
    """Class representing a graph of dependencies between correlation hooks."""

    def __init__(self, log):
        self.log = log.getChild("DependencyGraph")

        # dictionary of adjacency lists for each edge
        self._vertices = defaultdict(GraphVertex)
        self.topological_order = []

    def add_hook_dependency(self, hook_id: str, depends_on: list[str], may_change: list[str]):
        """Add hook to dependency graph and recalculate if any cycles are created."""
        if hook_id in self._vertices:
            raise ValueError(f"Hook id '{hook_id}' already present in the vertices.")
        for path in depends_on:
            self.add_edge(path, hook_id)
        for path in may_change:
            self.add_edge(hook_id, path)
        self._vertices[hook_id].type = "hook"
        try:
            self.topological_sort()
        except ValueError as err:
            raise ValueError(f"Hook {hook_id} introduces a circular dependency.") from err
        self.check_multiple_writes()

    def add_edge(self, id_from: Hashable, id_to: Hashable):
        """Add oriented edge between specified vertices."""
        self._vertices[id_from].adj.append(id_to)
        # Ensure vertex with 'id_to' exists to avoid iteration errors later.
        _ = self._vertices[id_to]

    def calculate_in_degrees(self):
        """Calculate number of incoming edges for each vertex. Time complexity O(V + E)."""
        for vertex_node in self._vertices.values():
            vertex_node.in_degree = 0

        for vertex_node in self._vertices.values():
            for adjacent_name in vertex_node.adj:
                self._vertices[adjacent_name].in_degree += 1

    def topological_sort(self):
        """
        Implementation of Kahn's algorithm for topological sorting.
        Raises ValueError if there is a cycle in the graph.

        See https://en.wikipedia.org/wiki/Topological_sorting#Kahn's_algorithm
        """
        self.calculate_in_degrees()
        queue = [(node_id, node) for node_id, node in self._vertices.items() if node.in_degree == 0]
        topological_order = []
        processed_vertices_cnt = 0

        while queue:
            curr_node_id, curr_node = queue.pop(0)
            topological_order.append(curr_node_id)

            # Decrease neighbouring nodes' in-degree by 1
            for neighbor in curr_node.adj:
                neighbor_node = self._vertices[neighbor]
                neighbor_node.in_degree -= 1
                # If in-degree becomes zero, add it to queue
                if neighbor_node.in_degree == 0:
                    queue.append((neighbor, neighbor_node))

            processed_vertices_cnt += 1

        if processed_vertices_cnt != len(self._vertices):
            raise ValueError("Dependency graph contains a cycle.")
        else:
            self.topological_order = topological_order
            return topological_order

    def check_multiple_writes(self):
        self.calculate_in_degrees()
        multiple_writes = [
            vertex_id
            for vertex_id, vertex in self._vertices.items()
            if vertex.in_degree > 1 and vertex.type == "attr"
        ]
        if multiple_writes:
            self.log.warning(
                "Detected possible over-write of hook results in attributes: %s", multiple_writes
            )



#!/usr/bin/env python3
"""Simple script to add min_t2s to master records to allow for better indexing.

Intended to be run only once to upgrade a database existing before 08-2024.
"""

import argparse
import time

from dp3.common.attrspec import AttrType
from dp3.common.config import ModelSpec, read_config_dir
from dp3.database.database import EntityDatabase, MongoConfig

# Arguments parser
parser = argparse.ArgumentParser(
    description="Add min_t2s to master records to allow for better indexing."
)
parser.add_argument(
    "--config",
    default="/etc/adict/config",
    help="DP3 config directory (default: /etc/adict/config)",
)
args = parser.parse_args()

# Load DP3 configuration
config = read_config_dir(args.config, recursive=True)
model_spec = ModelSpec(config.get("db_entities"))

# Connect to database
connection_conf = MongoConfig.model_validate(config.get("database", {}))
client = EntityDatabase.connect(connection_conf)
client.admin.command("ping")

db = client[connection_conf.db_name]


for entity, attributes in model_spec.entity_attributes.items():
    t1 = time.time()
    res = db[f"{entity}#master"].update_many(
        {},
        [
            {
                "$set": {
                    attr_name: {
                        "$cond": {
                            "if": {
                                "$eq": [
                                    {"$size": {"$ifNull": [f"${attr_name}", []]}},
                                    0,
                                ]
                            },
                            "then": "$$REMOVE",
                            "else": f"${attr_name}",
                        }
                    }
                    for attr_name, spec in attributes.items()
                    if spec.t != AttrType.PLAIN
                }
                | {
                    f"#min_t2s.{attr_name}": {
                        "$cond": {
                            "if": {
                                "$eq": [
                                    {"$size": {"$ifNull": [f"${attr_name}", []]}},
                                    0,
                                ]
                            },
                            "then": "$$REMOVE",
                            "else": {"$min": f"${attr_name}.t2"},
                        }
                    }
                    for attr_name, spec in attributes.items()
                    if spec.t != AttrType.PLAIN
                }
            },
        ],
    )
    t2 = time.time()
    print(f"Updated {res.modified_count} records for entity {entity} in {t2 - t1:.2f}s")



import datetime
import json
import logging
import os
import unittest
from functools import partial, update_wrapper
from typing import Callable, Optional

from event_count_logger import DummyEventGroup

from dp3.common.config import ModelSpec, PlatformConfig, read_config_dir
from dp3.common.task import Task
from dp3.snapshots.snapshooter import SnapShooter
from dp3.snapshots.snapshot_hooks import SnapshotCorrelationHookContainer


def modify_value(_: str, record: dict, attr: str, value):
    record[attr] = value


dummy_hook_abc = update_wrapper(partial(modify_value, attr="data2", value="abc"), modify_value)
dummy_hook_def = update_wrapper(partial(modify_value, attr="data1", value="def"), modify_value)


class TestHookDependency(unittest.TestCase):
    def setUp(self) -> None:
        config = read_config_dir(
            os.path.join(os.path.dirname(__file__), "..", "test_config"), recursive=True
        )
        self.model_spec = ModelSpec(config.get("db_entities"))
        self.container = SnapshotCorrelationHookContainer(
            log=logging.getLogger("TestLogger"), model_spec=self.model_spec, elog=DummyEventGroup()
        )

    def test_basic_function(self):
        self.container.register(
            hook=dummy_hook_abc, entity_type="A", depends_on=[["data1"]], may_change=[["data2"]]
        )
        values = {}
        self.container.run({("A", "a1"): values})
        self.assertEqual(values["data2"], "abc")

    def test_circular_dependency_error(self):
        self.container.register(
            hook=dummy_hook_abc, entity_type="A", depends_on=[["data1"]], may_change=[["data2"]]
        )
        with self.assertRaises(expected_exception=ValueError):
            self.container.register(
                hook=dummy_hook_def, entity_type="A", depends_on=[["data2"]], may_change=[["data1"]]
            )

    def validate_expanded_paths(self, hook_base, path, expected_expansions):
        hook_id = self.container.register(**hook_base, may_change=[path])
        vertex = self.container._dependency_graph._vertices[hook_id]
        for expected_expansion in expected_expansions:
            self.assertIn(expected_expansion, vertex.adj)
        self.assertEqual(len(expected_expansions), len(vertex.adj))

    def test_backlinks(self):
        hook_base = {"hook": dummy_hook_abc, "entity_type": "A", "depends_on": [["data2"]]}
        paths = [
            (["bs", "as", "bs", "data1"], ["B.data1"]),
            (["bs", "cs", "ds", "cs", "data1"], ["C.data1"]),
            (["bs", "as", "ds", "as", "data1"], ["A.data1"]),
        ]
        for path, expected_expansions in paths:
            with self.subTest(path=path, expected_expansions=expected_expansions):
                self.validate_expanded_paths(hook_base, path, expected_expansions)

    def test_nested_backlinks(self):
        hook_base = {"hook": dummy_hook_abc, "entity_type": "A", "depends_on": [["data2"]]}
        paths = [
            (["bs", "as", "bs", "data1"], ["B.data1"]),
            (["bs", "cs", "bs", "as", "data1"], ["A.data1"]),
        ]
        for path, expected_expansions in paths:
            with self.subTest(path=path, expected_expansions=expected_expansions):
                self.validate_expanded_paths(hook_base, path, expected_expansions)


class MockScheduler:
    def register(self, *_, **__):
        return None


class MockCursor(list):
    def close(self):
        return None


class MockDB:
    def __init__(self, content: dict, module_cache: dict):
        self.db_content = content
        self.module_cache = module_cache
        self.saved_snapshots = []
        self.saved_metadata = {}

    def get_master_records(self, etype: str, **kwargs):
        if etype not in self.db_content:
            return MockCursor([])
        return MockCursor(list(self.db_content[etype].values()))

    def get_master_record(self, etype: str, eid: str, projection=None) -> dict:
        if projection is not None:
            projection["_id"] = projection.get("_id", True)
            return {
                key: val
                for key, val in self.db_content[etype][eid].items()
                if key in projection and projection[key]
            }
        return self.db_content[etype][eid] or {}

    def register_on_entity_delete(
        self, f_one: Callable[[str, str], None], f_many: Callable[[str, list[str]], None]
    ): ...

    def get_module_cache(self, override_called_id: Optional[str] = None):
        return self.module_cache

    def save_snapshot(self, etype: str, snapshot: dict, time: datetime):
        json.dumps(snapshot)
        self.saved_snapshots.append(snapshot)

    def save_metadata(self, module: str, time: datetime, metadata: dict):
        self.saved_metadata[module, time] = metadata


class MockTaskQueueWriter:
    def __init__(self):
        self.tasks = []

    def put_task(self, task: Task, priority: bool = False):
        self.tasks.append(task)


class MockTaskQueueReader:
    def ack(self, msg_tag): ...


class MockCollection(dict):
    def create_index(self, *args, **kwargs): ...


class MockTaskExecutor:
    elog = DummyEventGroup()

    def register_attr_hook(self, *args, **kwargs): ...


class TestSnapshotOperation(unittest.TestCase):
    def setUp(self) -> None:
        config_base_path = os.path.join(os.path.dirname(__file__), "..", "test_config")
        config = read_config_dir(config_base_path, recursive=True)
        self.model_spec = ModelSpec(config.get("db_entities"))

        self.now = datetime.datetime.now()
        self.t1 = self.now - datetime.timedelta(minutes=30)
        self.t2 = self.now + datetime.timedelta(minutes=30)

        self.entities = {
            "B": {
                "b1": {
                    "_id": "b1",
                    "as": [{"v": "a1", "t1": self.t1, "t2": self.t2, "c": 1.0}],
                    "data1": {"v": "initb", "ts_last_update": self.now},
                    "data2": {"v": "initb", "ts_last_update": self.now},
                }
            },
            "A": {
                "a1": {
                    "_id": "a1",
                    "bs": [{"v": "b1", "t1": self.t1, "t2": self.t2, "c": 1.0}],
                    "data1": {"v": "inita", "ts_last_update": self.now},
                    "data2": {"v": "inita", "ts_last_update": self.now},
                }
            },
        }
        self.module_cache = MockCollection()
        self.db = MockDB(content=self.entities, module_cache=self.module_cache)

        self.snapshooter = SnapShooter(
            db=self.db,
            task_queue_writer=None,
            platform_config=PlatformConfig(
                app_name="test",
                config_base_path=config_base_path,
                config=config,
                model_spec=self.model_spec,
                process_index=0,
                num_processes=1,
            ),
            scheduler=MockScheduler(),
        )

    def test_multivalue_fitering(self):
        test_entity = {
            "_id": "t1",
            "test_attr_multi_value": [
                {"v": "a", "t1": self.t1, "t2": self.t2, "c": 1.0},
                {"v": "b", "t1": self.t1, "t2": self.t2, "c": 0.8},
                {"v": "c", "t1": self.t1, "t2": self.t2, "c": 0.4},
                {"v": "d", "t1": self.t1, "t2": self.t2, "c": 0.0},
            ],
        }

        values = self.snapshooter.get_values_at_time("test_entity_type", test_entity, self.now)
        self.assertEqual(len(values["test_attr_multi_value"]), 3)
        self.assertEqual(set(values["test_attr_multi_value"]), {"a", "b", "c"})


if __name__ == "__main__":
    unittest.main()



"""
[SnapShooter][dp3.snapshots.snapshooter], a module responsible
for snapshot creation and running configured data correlation and fusion hooks,
and [Snapshot Hooks][dp3.snapshots.snapshot_hooks], which manage the registered hooks and their
dependencies on one another.
"""



from datetime import datetime, timedelta
from typing import Annotated, Union

from event_count_logger import DummyEventGroup, EventGroup
from pydantic import AfterValidator, BeforeValidator
from pydantic_core.core_schema import FieldValidationInfo

from dp3.common.utils import parse_time_duration, time_duration_pattern


def parse_timedelta_or_passthrough(v):
    """
    We pass the value to the native pydantic validator if the value does not match our pattern.
    """
    if v and isinstance(v, str) and time_duration_pattern.match(v):
        return parse_time_duration(v)
    return v


ParsedTimedelta = Annotated[timedelta, BeforeValidator(parse_timedelta_or_passthrough)]


def t2_implicity_t1(v, info: FieldValidationInfo):
    """If t2 is not specified, it is set to t1."""
    v = v or info.data.get("t1")
    return v


def t2_after_t1(v, info: FieldValidationInfo):
    """t2 must be after t1"""
    if info.data.get("t1"):
        assert info.data["t1"] <= v, "'t2' is before 't1'"
    return v


T2Datetime = Annotated[datetime, BeforeValidator(t2_implicity_t1), AfterValidator(t2_after_t1)]

EventGroupType = Union[EventGroup, DummyEventGroup]



import logging
from typing import Callable

from dp3.common.attrspec import AttrType
from dp3.common.config import ModelSpec
from dp3.common.datapoint import DataPointBase
from dp3.common.task import DataPointTask, task_context
from dp3.common.types import EventGroupType
from dp3.common.utils import get_func_name


class TaskGenericHooksContainer:
    """Container for generic hooks

    Possible hooks:

    - `on_task_start`: receives Task, no return value requirements
    """

    def __init__(self, log: logging.Logger, elog: EventGroupType):
        self.log = log.getChild("genericHooks")
        self.elog = elog

        self._on_start = []

    def register(self, hook_type: str, hook: Callable):
        if hook_type == "on_task_start":
            self._on_start.append(hook)
        else:
            raise ValueError(f"Hook type '{hook_type}' doesn't exist.")

        self.log.debug(f"Added '{hook_type}' hook: {get_func_name(hook)}")

    def run_on_start(self, task: DataPointTask):
        for hook in self._on_start:
            # Run hook
            try:
                hook(task)
            except Exception as e:
                self.elog.log("module_error")
                self.log.error(f"Error during running hook {hook}: {e}")


class TaskEntityHooksContainer:
    """Container for entity hooks

    Possible hooks:

    - `allow_entity_creation`: receives eid and Task, may prevent entity record creation (by
          returning False)
    - `on_entity_creation`: receives eid and Task, may return list of DataPointTasks
    """

    def __init__(
        self, entity: str, model_spec: ModelSpec, log: logging.Logger, elog: EventGroupType
    ):
        self.entity = entity
        self.log = log.getChild(f"entityHooks.{entity}")
        self.elog = elog
        self.model_spec = model_spec

        self._allow_creation = []
        self._on_creation = []

    def register(self, hook_type: str, hook: Callable):
        if hook_type == "allow_entity_creation":
            self._allow_creation.append(hook)
        elif hook_type == "on_entity_creation":
            self._on_creation.append(hook)
        else:
            raise ValueError(f"Hook type '{hook_type}' doesn't exist.")

        self.log.debug(f"Added '{hook_type}' hook: {get_func_name(hook)}")

    def run_allow_creation(self, eid: str, task: DataPointTask):
        for hook in self._allow_creation:
            try:
                if not hook(eid, task):
                    self.log.debug(
                        f"Creation of eid '{eid}' prevented because hook '{hook}' returned False."
                    )
                    return False
            except Exception as e:
                self.elog.log("module_error")
                self.log.error(f"Error during running hook {get_func_name(hook)}: {e}")

        return True

    def run_on_creation(self, eid: str, task: DataPointTask):
        new_tasks = []

        with task_context(self.model_spec):
            for hook in self._on_creation:
                try:
                    # Run hook
                    hook_new_tasks = hook(eid, task)

                    # Append new tasks to process
                    if isinstance(hook_new_tasks, list):
                        new_tasks += hook_new_tasks
                except Exception as e:
                    self.elog.log("module_error")
                    self.log.error(f"Error during running hook {hook}: {e}")

        return new_tasks


class TaskAttrHooksContainer:
    """Container for attribute hooks

    Possible hooks:

    - `on_new_plain`, `on_new_observation`, `on_new_ts_chunk`:
        receives eid and DataPointBase, may return a list of DataPointTasks
    """

    def __init__(
        self,
        entity: str,
        attr: str,
        attr_type: AttrType,
        model_spec: ModelSpec,
        log: logging.Logger,
        elog: EventGroupType,
    ):
        self.entity = entity
        self.attr = attr
        self.log = log.getChild(f"attributeHooks.{entity}.{attr}")
        self.elog = elog
        self.model_spec = model_spec

        if attr_type == AttrType.PLAIN:
            self.on_new_hook_type = "on_new_plain"
        elif attr_type == AttrType.OBSERVATIONS:
            self.on_new_hook_type = "on_new_observation"
        elif attr_type == AttrType.TIMESERIES:
            self.on_new_hook_type = "on_new_ts_chunk"
        else:
            raise ValueError(f"Invalid attribute type '{attr_type}'")

        self._on_new = []

    def register(self, hook_type: str, hook: Callable):
        if hook_type == self.on_new_hook_type:
            self._on_new.append(hook)
        else:
            raise ValueError(
                f"Hook type '{hook_type}' doesn't exist for {self.entity}/{self.attr}."
            )

        self.log.debug(f"Added '{hook_type}' hook: {get_func_name(hook)}")

    def run_on_new(self, eid: str, dp: DataPointBase):
        new_tasks = []

        with task_context(self.model_spec):
            for hook in self._on_new:
                try:
                    # Run hook
                    hook_new_tasks = hook(eid, dp)

                    # Append new tasks to process
                    if isinstance(hook_new_tasks, list):
                        new_tasks += hook_new_tasks
                except Exception as e:
                    self.elog.log("module_error")
                    self.log.error(f"Error during running hook {hook}: {e}")

        return new_tasks



"""
Load and check configuration from given directory, print any errors, and exit.

TODO:
 - refactor to simplify the code, some error path matching must be done to counteract
   the AttrSpec function magic where Pydantic fails, but otherwise it is not required
"""

import argparse
import enum
import json
import sys
from datetime import datetime
from json import JSONEncoder
from typing import Any, get_args, get_origin, get_type_hints

import yaml
from pydantic import BaseModel, ValidationError

from dp3.common.attrspec import AttrSpec
from dp3.common.config import EntitySpecDict, ModelSpec, read_config_dir

special_model_cases = {
    (EntitySpecDict, "attribs"): AttrSpec,
}


class ConfigEncoder(JSONEncoder):
    """JSONEncoder to encode parsed configuration."""

    def default(self, o: Any) -> Any:
        if isinstance(o, BaseModel):
            return o.model_dump(exclude_none=True)
        if isinstance(o, enum.Enum):
            return o.value
        if isinstance(o, datetime):
            return o.strftime("%Y-%m-%dT%H:%M:%S.%fZ")[:-4]
        return str(o)


def stringify_source(source) -> str:
    if source is None:
        return ""

    if isinstance(source, dict):
        source_str: str = yaml.dump(source, default_flow_style=False, width=1000)
        source_str = "```\n  " + source_str.replace("\n", "\n  ")[:-2] + "```"
    else:
        source_str = f"\t`{source}`"
    return source_str


def get_all_attribs_errors(data: dict, error: dict) -> tuple[list[tuple[str]], list]:
    paths, sources = locate_attribs_error(data, error["msg"])

    path: list[str] = list(error["loc"])
    paths = [tuple(path[:3] + list(p)) for p in paths]

    return paths, sources


def locate_attribs_error(data: dict, sought_err: str) -> tuple[list[tuple], list[dict]]:
    """
    Locate source of an error in a dict of AttrSpecs.
    Returns all sources of the same kind of error.
    """
    paths = []
    sources = []

    for attr, attr_spec in data.items():
        try:
            AttrSpec(attr, attr_spec)
        except ValidationError as exception:
            for err_dict in exception.errors():
                if err_dict["msg"] == sought_err:
                    paths.append((attr, *err_dict["loc"]))
                    sources.append(attr_spec)
        except (ValueError, AssertionError) as exception:
            if sought_err.endswith(exception.args[0]):
                paths.append((attr,))
                sources.append(attr_spec)

    return paths, sources


def locate_basemodel_error(data: list[dict], model: BaseModel) -> list[dict]:
    locations = []
    if isinstance(data, list):
        for item in data:
            try:
                model(item)
            except (ValidationError, ValueError, AssertionError):
                locations.append(item)
    return locations


def get_error_sources(data: dict, error: dict) -> tuple[list, list]:
    """Locate source of an error in validated data using Pydantic error path."""
    err_path = error["loc"]

    # Kickstart the model exploration
    curr_model_dict = get_type_hints(ModelSpec)["config"]
    curr_model_origin = get_origin(curr_model_dict)
    curr_model = get_args(curr_model_dict)[1]

    for key in err_path[1:]:
        if curr_model_origin is not dict and curr_model_origin is not None:
            return [err_path], [data]

        if key in data:
            prev_data = data
            data = data[key]

            if (curr_model, key) in special_model_cases:
                curr_model = special_model_cases[curr_model, key]
                curr_model_dict = get_type_hints(curr_model)
                curr_model_origin = get_origin(curr_model_dict)

                if curr_model == AttrSpec:
                    return get_all_attribs_errors(data, error)
                continue

            if isinstance(curr_model_dict, dict):
                if key in curr_model_dict:
                    curr_model = curr_model_dict[key]
                else:
                    return [err_path], [prev_data]

            curr_model_dict = get_type_hints(curr_model)
            curr_model_origin = get_origin(curr_model_dict)
            if curr_model_origin is dict:
                curr_model = get_args(curr_model_dict)[1]
        else:
            return [err_path], [data]

    if curr_model == AttrSpec:
        return get_all_attribs_errors(data, error)

    return [], locate_basemodel_error(data, curr_model)


def locate_errors(exc: ValidationError, data: dict):
    """Locate errors (i.e.: get the paths and sources) in a ValidationError object."""
    paths = []
    sources = []
    errors = []

    for error in exc.errors():
        if error["loc"] == ():
            paths.append(())
            sources.append(None)
            errors.append(error["msg"])
            continue

        message = f'{error["msg"]} (type={error["type"]})'
        e_paths, e_sources = get_error_sources(data, error)

        paths.extend(e_paths)
        sources.extend(e_sources)
        errors.extend(message for _ in range(len(e_paths)))

    return paths, sources, errors


def init_parser(parser):
    parser.add_argument(
        "config_dir",
        metavar="CONFIG_DIRECTORY",
        help="Path to a directory containing configuration files (e.g. /etc/my_app/config)",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Verbose mode - print parsed configuration",
        default=False,
    )


def run():
    print(
        "WARNING: The `check` entrypoint is deprecated due to possible namespace conflicts. "
        "Please use `dp3 check` instead.",
        file=sys.stderr,
    )

    # Parse arguments
    parser = argparse.ArgumentParser(
        prog="check",
        description="Load configuration from given directory and check its validity. "
        "When configuration is OK, program exits immediately with status code 0, "
        "otherwise it prints error messages on stderr and exits with non-zero status.",
    )
    init_parser(parser)
    args = parser.parse_args()

    main(args)


def main(args):
    try:
        config = read_config_dir(args.config_dir, recursive=True)
    except OSError:
        print(f"Cannot open: {args.config_dir}", sys.stderr)
        sys.exit(1)

    assert "db_entities" in config, "Config for 'db_entities' is missing"
    assert "database" in config, "Config for 'database' is missing"

    db_config = config.get("db_entities")
    try:
        parsed_config = ModelSpec(db_config)
    except ValidationError as exc:
        print("Invalid model specification (check entity config):")
        paths, sources, errors = locate_errors(exc, db_config)

        # Print each source only once
        unique_sources = []
        source_paths_and_errors = []

        for path, source, err in zip(paths, sources, errors):
            if source in unique_sources:
                i = unique_sources.index(source)
                source_paths_and_errors[i].add((path, err))
            else:
                unique_sources.append(source)
                source_paths_and_errors.append({(path, err)})

        for source, paths_and_errors in zip(unique_sources, source_paths_and_errors):
            for path, err in paths_and_errors:
                print(" -> ".join(path))
                print("  ", err)
            print(stringify_source(source), "\n")
        sys.exit(1)

    if args.verbose:
        # Print parsed config as JSON (print unserializable objects using str())
        print(json.dumps(parsed_config.config, indent=4, cls=ConfigEncoder))
    sys.exit(0)


if __name__ == "__main__":
    run()



import logging

from fastapi import FastAPI
from fastapi.exception_handlers import request_validation_exception_handler
from fastapi.exceptions import RequestValidationError
from fastapi.middleware.cors import CORSMiddleware

from dp3.api.internal.config import CONFIG, DATAPOINTS_INGESTION_URL_PATH, DP_LOGGER, ROOT_PATH
from dp3.api.routers import control, entity, root, telemetry

uvicorn_logger = logging.getLogger("uvicorn")
uvicorn_access_logger = logging.getLogger("uvicorn.access")
gunicorn_error_logger = logging.getLogger("gunicorn.error")

if len(gunicorn_error_logger.handlers) > 0:  # We are running in gunicorn
    logging.root.handlers = gunicorn_error_logger.handlers
    uvicorn_access_logger.handlers = gunicorn_error_logger.handlers
elif len(uvicorn_logger.handlers) > 0:  # We are running only in uvicorn
    logging.root.handlers = uvicorn_logger.handlers

# Create new FastAPI app
app = FastAPI(root_path=ROOT_PATH)


# Redefine Pydantic validation error handler to log all bad datapoints
@app.exception_handler(RequestValidationError)
async def validation_exception_handler(request, exc):
    if request.url.path.replace(ROOT_PATH, "") == DATAPOINTS_INGESTION_URL_PATH:
        # Log individual errors or whole exception
        for err in exc.errors():
            # Override empty input on json decode error with request data
            if err["type"] == "json_invalid":
                err["input"] = str(exc.body)
            DP_LOGGER.log_bad(str(err), src=request.client.host)
        if not exc.errors():
            DP_LOGGER.log_bad(str(exc), src=request.client.host)

    # Call default handler
    return await request_validation_exception_handler(request, exc)


# CORS
app.add_middleware(
    CORSMiddleware,
    allow_origins=CONFIG.get("api.cors.allow_origins", ["*"]),
    allow_methods=["*"],
    allow_headers=["*"],
)

# Register routers
app.include_router(entity.router, prefix="/entity", tags=["Entity"])
app.include_router(control.router, prefix="/control", tags=["Control"])
app.include_router(telemetry.router, prefix="/telemetry", tags=["Telemetry"])
app.include_router(root.router)



import logging
from functools import partial
from logging import Logger
from typing import Callable, Union

from pydantic import BaseModel

from dp3.common.attrspec import AttrType
from dp3.common.config import ModelSpec, PlatformConfig, read_config_dir
from dp3.common.datapoint import DataPointType
from dp3.common.scheduler import Scheduler
from dp3.common.state import SharedFlag
from dp3.common.task import DataPointTask
from dp3.common.types import ParsedTimedelta
from dp3.core.updater import Updater
from dp3.snapshots.snapshooter import SnapShooter
from dp3.task_processing.task_executor import TaskExecutor


def write_datapoints_into_record(model_spec: ModelSpec, tasks: list[DataPointTask], record: dict):
    eid = record["eid"]
    for task in tasks:
        if task.eid != eid:
            continue
        for dp in task.data_points:
            attr_spec = model_spec.attributes[dp.etype, dp.attr]
            if attr_spec.t == AttrType.PLAIN:
                if attr_spec.t in AttrType.PLAIN | AttrType.OBSERVATIONS and attr_spec.is_iterable:
                    v = [
                        elem.model_dump() if isinstance(elem, BaseModel) else elem for elem in dp.v
                    ]
                else:
                    v = dp.v.model_dump() if isinstance(dp.v, BaseModel) else dp.v

                record[dp.attr] = v


def on_entity_creation_in_snapshots(
    model_spec: ModelSpec,
    run_flag: SharedFlag,
    original_hook: Callable[[str, DataPointTask], list[DataPointTask]],
    etype: str,
    record: dict,
) -> list[DataPointTask]:
    """Wrapper for on_entity_creation hooks to enable running as a snapshot callback."""
    if not run_flag.isset():
        return []
    eid = record["eid"]
    mock_task = DataPointTask(etype=etype, eid=eid, data_points=[])
    tasks = original_hook(eid, mock_task)
    write_datapoints_into_record(model_spec, tasks, record)
    return tasks


def on_attr_change_in_snapshots(
    model_spec: ModelSpec,
    run_flag: SharedFlag,
    original_hook: Callable[[str, DataPointTask], Union[list[DataPointTask], None]],
    etype: str,
    record: dict,
) -> list[DataPointTask]:
    """Wrapper for on_entity_creation hooks to enable running as a snapshot callback."""
    if not run_flag.isset():
        return []
    eid = record["eid"]
    mock_task = DataPointTask(etype=etype, eid=eid, data_points=[])
    tasks = original_hook(eid, mock_task)
    if isinstance(tasks, list):
        write_datapoints_into_record(model_spec, tasks, record)
    return tasks


def unset_flag(flag: SharedFlag) -> list:
    flag.unset()
    return []


def reload_module_config(
    log: Logger, platform_config: PlatformConfig, modules: dict, module: str
) -> None:
    """
    Reloads configuration of a module.

    Args:
        log: log to write messages to
        platform_config: Platform configuration
        modules: Dictionary of loaded modules by their names
        module: Name of the module to reload

    Returns:
        Module's configuration
    """
    log.debug(f"Reloading config of module '{module}'")

    if module not in modules:
        log.warning(f"Could not find module '{module}', cannot reload config. Is it loaded?")
        return

    config = read_config_dir(platform_config.config_base_path, recursive=True)
    module_config = config.get(f"modules.{module}", {})
    modules[module].load_config(platform_config, module_config)
    modules[module].refresh.set()
    log.info(f"Config of module '{module}' reloaded successfully.")


class CallbackRegistrar:
    """Interface for callback registration."""

    def __init__(
        self,
        scheduler: Scheduler,
        task_executor: TaskExecutor,
        snap_shooter: SnapShooter,
        updater: Updater,
    ):
        self._scheduler = scheduler
        self._task_executor = task_executor
        self._snap_shooter = snap_shooter
        self._updater = updater

        self.log = logging.getLogger(self.__class__.__name__)
        self.model_spec = task_executor.model_spec
        self.attr_spec_t_to_on_attr = {
            AttrType.PLAIN: "on_new_plain",
            AttrType.OBSERVATIONS: "on_new_observation",
            AttrType.TIMESERIES: "on_new_ts_chunk",
        }

    def scheduler_register(
        self,
        func: Callable,
        *,
        func_args: Union[list, tuple] = None,
        func_kwargs: dict = None,
        year: Union[int, str] = None,
        month: Union[int, str] = None,
        day: Union[int, str] = None,
        week: Union[int, str] = None,
        day_of_week: Union[int, str] = None,
        hour: Union[int, str] = None,
        minute: Union[int, str] = None,
        second: Union[int, str] = None,
        timezone: str = "UTC",
        misfire_grace_time: int = 1,
    ) -> int:
        """
        Register a function to be run at specified times.

        Pass cron-like specification of when the function should be called,
        see [docs](https://apscheduler.readthedocs.io/en/latest/modules/triggers/cron.html)
        of apscheduler.triggers.cron for details.

        Args:
            func: function or method to be called
            func_args: list of positional arguments to call func with
            func_kwargs: dict of keyword arguments to call func with
            year: 4-digit year
            month: month (1-12)
            day: day of month (1-31)
            week: ISO week (1-53)
            day_of_week: number or name of weekday (0-6 or mon,tue,wed,thu,fri,sat,sun)
            hour: hour (0-23)
            minute: minute (0-59)
            second: second (0-59)
            timezone: Timezone for time specification (default is UTC).
            misfire_grace_time: seconds after the designated run time
                that the job is still allowed to be run (default is 1)
        Returns:
             job ID
        """
        return self._scheduler.register(
            func,
            func_args=func_args,
            func_kwargs=func_kwargs,
            year=year,
            month=month,
            day=day,
            week=week,
            day_of_week=day_of_week,
            hour=hour,
            minute=minute,
            second=second,
            timezone=timezone,
            misfire_grace_time=misfire_grace_time,
        )

    def register_task_hook(self, hook_type: str, hook: Callable):
        """Registers one of available task hooks

        See: [`TaskGenericHooksContainer`][dp3.task_processing.task_hooks.TaskGenericHooksContainer]
        in `task_hooks.py`
        """
        self._task_executor.register_task_hook(hook_type, hook)

    def register_allow_entity_creation_hook(
        self,
        hook: Callable[[str, DataPointTask], bool],
        entity: str,
    ):
        """
        Registers passed hook to allow entity creation.

        Binds hook to specified entity (though same hook can be bound multiple times).

        Args:
            hook: `hook` callable should expect eid and Task as arguments and return a bool.
            entity: specifies entity type
        """
        self._task_executor.register_entity_hook("allow_entity_creation", hook, entity)

    def register_on_entity_creation_hook(
        self,
        hook: Callable[[str, DataPointTask], list[DataPointTask]],
        entity: str,
        refresh: SharedFlag = None,
        may_change: list[list[str]] = None,
    ):
        """
        Registers passed hook to be called on entity creation.

        Binds hook to specified entity (though same hook can be bound multiple times).

        Allows registration of refreshing on configuration changes, if `refresh` is specified.
        In that case, `may_change` must be specified.

        Args:
            hook: `hook` callable should expect eid and Task as arguments and
                return a list of DataPointTask objects to perform.
            entity: specifies entity type
            refresh: If specified, registered hook will be called on configuration changes.
                Pass `self.refresh` from `BaseModule` subclasses.
            may_change: each item should specify an attribute that `hook` may change,
                for specification format see `register_correlation_hook`
        """
        self._task_executor.register_entity_hook("on_entity_creation", hook, entity)
        if refresh is not None:
            if may_change is None:
                raise ValueError("'may_change' must be specified if 'refresh' is specified")
            self._snap_shooter.register_correlation_hook(
                partial(on_entity_creation_in_snapshots, self.model_spec, refresh, hook),
                entity,
                [],
                may_change,
            )
            self._snap_shooter.register_run_finalize_hook(partial(unset_flag, refresh))

    def register_entity_hook(self, hook_type: str, hook: Callable, entity: str):
        """Registers one of available task entity hooks

        !!! warning "Deprecated"

            This method is deprecated, use `register_on_entity_creation_hook`
            or `register_allow_entity_creation_hook` instead.

        See: [`TaskEntityHooksContainer`][dp3.task_processing.task_hooks.TaskEntityHooksContainer]
        in `task_hooks.py`
        """
        self._task_executor.register_entity_hook(hook_type, hook, entity)

    def register_on_new_attr_hook(
        self,
        hook: Callable[[str, DataPointType], Union[None, list[DataPointTask]]],
        entity: str,
        attr: str,
        refresh: SharedFlag = None,
        may_change: list[list[str]] = None,
    ):
        """
        Registers passed hook to be called on new attribute datapoint.

        Args:
            hook: `hook` callable should expect eid, Task and attribute value as arguments.
                Can optionally return a list of DataPointTasks to perform.
            entity: specifies entity type
            attr: specifies attribute name
            refresh: If specified, registered hook will be called on configuration changes.
                Pass `self.refresh` from `BaseModule` subclasses.
            may_change: each item should specify an attribute that `hook` may change,
                for specification format see `register_correlation_hook`

        Raises:
            ValueError: If entity and attr do not specify a valid attribute, a ValueError is raised.
        """
        try:
            hook_type = self.attr_spec_t_to_on_attr[self.model_spec.attributes[entity, attr].t]
        except KeyError as e:
            raise ValueError(
                f"Cannot register hook for attribute {entity}/{attr}, are you sure it exists?"
            ) from e
        self._task_executor.register_attr_hook(hook_type, hook, entity, attr)

        if refresh is None:
            return
        if may_change is None:
            raise ValueError("'may_change' must be specified if 'refresh' is specified")

        self._snap_shooter.register_correlation_hook(
            partial(on_attr_change_in_snapshots, self.model_spec, refresh, hook),
            entity,
            [[attr]],
            may_change,
        )
        self._snap_shooter.register_run_finalize_hook(partial(unset_flag, refresh))

    def register_attr_hook(self, hook_type: str, hook: Callable, entity: str, attr: str):
        """
        Registers one of available task attribute hooks

        !!! warning "Deprecated"

            This method is deprecated, use `register_on_new_attr_hook` instead.

        See: [`TaskAttrHooksContainer`][dp3.task_processing.task_hooks.TaskAttrHooksContainer]
        in `task_hooks.py`
        """
        self._task_executor.register_attr_hook(hook_type, hook, entity, attr)

    def register_timeseries_hook(
        self,
        hook: Callable[[str, str, list[dict]], list[DataPointTask]],
        entity_type: str,
        attr_type: str,
    ):
        """
        Registers passed timeseries hook to be called during snapshot creation.

        Binds hook to specified `entity_type` and `attr_type` (though same hook can be bound
        multiple times).

        Args:
            hook: `hook` callable should expect entity_type, attr_type and attribute
                history as arguments and return a list of `DataPointTask` objects.
            entity_type: specifies entity type
            attr_type: specifies attribute type

        Raises:
            ValueError: If entity_type and attr_type do not specify a valid timeseries attribute,
                a ValueError is raised.
        """
        self._snap_shooter.register_timeseries_hook(hook, entity_type, attr_type)

    def register_correlation_hook(
        self,
        hook: Callable[[str, dict], Union[None, list[DataPointTask]]],
        entity_type: str,
        depends_on: list[list[str]],
        may_change: list[list[str]],
    ):
        """
        Registers passed hook to be called during snapshot creation.

        Binds hook to specified entity_type (though same hook can be bound multiple times).

        `entity_type` and attribute specifications are validated, `ValueError` is raised on failure.

        Args:
            hook: `hook` callable should expect entity type as str
                and its current values, including linked entities, as dict.
                Can optionally return a list of DataPointTask objects to perform.
            entity_type: specifies entity type
            depends_on: each item should specify an attribute that is depended on
                in the form of a path from the specified entity_type to individual attributes
                (even on linked entities).
            may_change: each item should specify an attribute that `hook` may change.
                specification format is identical to `depends_on`.

        Raises:
            ValueError: On failure of specification validation.
        """
        self._snap_shooter.register_correlation_hook(hook, entity_type, depends_on, may_change)

    def register_snapshot_init_hook(self, hook: Callable[[], list[DataPointTask]]):
        """
        Registers passed hook to be called before a run of snapshot creation begins.

        Args:
            hook: `hook` callable should expect no arguments and
                return a list of DataPointTask objects to perform.
        """
        self._snap_shooter.register_run_init_hook(hook)

    def register_snapshot_finalize_hook(self, hook: Callable[[], list[DataPointTask]]):
        """
        Registers passed hook to be called after a run of snapshot creation ends.

        Args:
            hook: `hook` callable should expect no arguments and
                return a list of DataPointTask objects to perform.
        """
        self._snap_shooter.register_run_finalize_hook(hook)

    def register_periodic_update_hook(
        self,
        hook: Callable[[str, str, dict], list[DataPointTask]],
        hook_id: str,
        entity_type: str,
        period: ParsedTimedelta,
    ):
        """
        Registers a callback for periodic update of entities of the specified type.

        The callback receives the entity type, the entity ID and the master record.

        Args:
            hook: `hook` callable should expect entity type, entity ID and master record
                as arguments and return a list of DataPointTask objects to perform.
            hook_id: specifies hook ID
            entity_type: specifies entity type
            period: specifies period of the callback
        """
        self._updater.register_record_update_hook(hook, hook_id, entity_type, period)

    def register_periodic_eid_update_hook(
        self,
        hook: Callable[[str, str], list[DataPointTask]],
        hook_id: str,
        entity_type: str,
        period: ParsedTimedelta,
    ):
        """
        Registers a callback for periodic update of entities of the specified type.

        The callback receives the entity type and the entity ID.

        Args:
            hook: `hook` callable should expect entity type and entity ID as arguments
                and return a list of DataPointTask objects to perform.
            hook_id: specifies hook ID
            entity_type: specifies entity type
            period: specifies period of the callback
        """
        self._updater.register_eid_update_hook(hook, hook_id, entity_type, period)



import logging
import os
import sys
import time
import unittest
from typing import Callable, TypeVar

import requests
from pydantic import BaseModel

from dp3.common.config import ModelSpec, read_config_dir

Model = TypeVar("Model", bound=BaseModel)  # Can be any subtype of BaseModel
A = TypeVar("A")

base_url = os.getenv("BASE_URL", default="http://127.0.0.1:5000")
RECONNECT_DELAYS = [1, 2, 5, 10, 30]

ACCEPTED_ERROR_CODES = {400, 422}


class ConfigEnv(BaseModel):
    """Configuration environment variables container"""

    CONF_DIR: str


conf_env = ConfigEnv.parse_obj(os.environ)
CONFIG = read_config_dir(conf_env.CONF_DIR, recursive=True)
MODEL_SPEC = ModelSpec(CONFIG.get("db_entities"))

values = {
    "valid": {
        "binary": ["true"],
        "int": ["123"],
        "int64": ["123"],
        "string": ['"xyz"'],
        "float": ["1.0", "1"],
        "ipv4": ["127.0.0.1"],
        "ipv6": ["2001:0db8:85a3:0000:0000:8a2e:0370:7334", "::1"],
        "mac": ["de:ad:be:ef:ba:be", "11:22:33:44:55:66"],
        "time": ["2020-01-01T00:00:00"],
        "json": ['{"test": "test"}'],
        "category": ["cat1"],
        "array": [[1, 2, 3]],
        "set": [[1, 2, 3]],
        "dict": [{"key1": 1, "key2": "xyz"}],
        "link": [{"eid": "test_entity_2"}],
        "data_link": [{"eid": "test_entity_3", "data": 42}],
    },
    "invalid": {
        "binary": ["xyz"],
        "int": ["xyz"],
        "int64": ["xyz"],
        "string": [],  # all JSON strings can be converted to strings
        "float": ["xyz"],
        "ipv4": ['"xyz"'],
        "ipv6": ['"xyz"'],
        "mac": ['"xyz"'],
        "time": ['"xyz"'],
        "json": ["xyz"],
        "category": ["xyz"],
        "array": ["xyz", '["xyz"]'],
        "set": ["xyz", '["xyz"]'],
        "dict": ["xyz", '{"xyz":"xyz"}', '{"key1":"xyz","key2":"xyz"}'],
        "link": ["test_entity_2", {"id": "test_entity_2"}],
        "data_link": [
            "test_entity_2",
            {"id": "test_entity_2"},
            {"eid": "test_entity_3"},
            {"eid": "test_entity_3", "data": [42]},
        ],
    },
}


def retry_request_on_error(request):
    for attempt, delay in enumerate(RECONNECT_DELAYS):
        try:
            return request()
        except (requests.exceptions.ConnectionError, requests.exceptions.ReadTimeout) as err:
            logging.warning("Connection failed, retrying in %ds (attempt %d)", delay, attempt + 1)
            time.sleep(delay)
            if attempt + 1 == len(RECONNECT_DELAYS):
                raise err


class APITest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        # Test the base endpoint is live before running tests.
        try:
            retry_request_on_error(lambda: requests.get(base_url))
            api_up = True
        except requests.exceptions.ConnectionError:
            api_up = False
        assert api_up, "API is down."

    @staticmethod
    def post_request(path, **kwargs) -> requests.Response:
        return retry_request_on_error(
            lambda: requests.post(f"{base_url}/{path}", **kwargs, timeout=5)
        )

    def query_expected_value(
        self, query: Callable[[], A], assertion: Callable[[A], bool], attempts=5, delay_s=0.5
    ):
        payload: A = None
        for _ in range(attempts):
            time.sleep(delay_s)
            payload = query()
            print(payload, file=sys.stderr)
            if assertion(payload):
                break
        self.assertTrue(assertion(payload))
        return payload

    @classmethod
    def push_datapoints(cls, json_data) -> requests.Response:
        return cls.post_request("datapoints", json=json_data)

    def get_entity_data(self, path: str, model: type[Model], **kwargs) -> Model:
        response = self.get_request(path, **kwargs)
        self.assertEqual(response.status_code, 200)
        return model.model_validate_json(response.content)

    @staticmethod
    def get_request(path, **kwargs) -> requests.Response:
        args_str = "&".join([f"{key}={value}" for key, value in kwargs.items()])
        if args_str:
            args_str = f"?{args_str}"
        return retry_request_on_error(
            lambda: requests.get(f"{base_url}/{path}{args_str}", timeout=5)
        )



"""
Module enabling remote control of the platform's internal events.
"""

import logging
from enum import Enum
from typing import Callable

from pydantic import BaseModel

from dp3.common.config import PlatformConfig
from dp3.common.task import Task
from dp3.common.utils import get_func_name
from dp3.task_processing.task_distributor import TaskDistributor
from dp3.task_processing.task_executor import TaskExecutor
from dp3.task_processing.task_queue import TaskQueueReader


class ControlAction(Enum):
    make_snapshots = "make_snapshots"
    refresh_on_entity_creation = "refresh_on_entity_creation"
    refresh_module_config = "refresh_module_config"


class ControlConfig(BaseModel):
    allowed_actions: list[ControlAction]


class ControlMessage(Task):
    action: ControlAction
    kwargs: dict = {}

    def routing_key(self):
        return ""

    def hashed_routing_key(self) -> int:
        return 0

    def as_message(self) -> str:
        return self.model_dump_json()


class Control:
    """Class enabling remote control of the platform's internal events."""

    def __init__(
        self,
        platform_config: PlatformConfig,
    ) -> None:
        self.log = logging.getLogger("Control")
        self.action_handlers: dict[ControlAction, Callable] = {}

        self.config = ControlConfig.model_validate(platform_config.config.get("control"))
        self.allowed_actions = set(self.config.allowed_actions)
        self.log.debug("Allowed actions: %s", self.allowed_actions)

        queue = f"{platform_config.app_name}-worker-{platform_config.process_index}-control"
        self.control_queue = TaskQueueReader(
            callback=self.process_control_task,
            parse_task=ControlMessage.model_validate_json,
            app_name=platform_config.app_name,
            worker_index=platform_config.process_index,
            rabbit_config=platform_config.config.get("processing_core.msg_broker", {}),
            queue=queue,
            priority_queue=False,
            parent_logger=self.log,
        )

    def start(self):
        """Connect to RabbitMQ and start consuming from TaskQueue."""
        unconfigured_handlers = self.allowed_actions - set(self.action_handlers)
        if unconfigured_handlers:
            raise ValueError(
                f"The following configured actions are missing handlers: {unconfigured_handlers}"
            )

        self.log.info("Connecting to RabbitMQ")
        self.control_queue.connect()
        self.control_queue.check()  # check presence of needed queues
        self.control_queue.start()

        self.log.debug(
            "Configured handlers: %s", ", ".join(get_func_name(f) for f in self.action_handlers)
        )

    def stop(self):
        """Stop consuming from TaskQueue, disconnect from RabbitMQ."""
        self.control_queue.stop()
        self.control_queue.disconnect()

    def set_action_handler(self, action: ControlAction, handler: Callable):
        """Sets the handler for the given action"""
        self.log.debug("Setting handler for action %s: %s", action, get_func_name(handler))
        self.action_handlers[action] = handler

    def process_control_task(self, msg_id, task: ControlMessage):
        """
        Acknowledges the received message and executes an action according to the `task`.

        This function should not be called directly, but set as callback for TaskQueueReader.
        """
        if not self.control_queue.ack(msg_id):
            return
        if task.action in self.allowed_actions:
            self.log.info("Executing action: %s", task.action)
            self.action_handlers[task.action](**task.kwargs)
            self.log.info("Action finished: %s", task.action)
        else:
            self.log.error("Action not allowed: %s", task.action)


def refresh_on_entity_creation(
    task_distributor: TaskDistributor, task_executor: TaskExecutor, etype: str
):
    """Refreshes hooks called on new entity creation for all entities in DB."""
    tasks = task_executor.refresh_on_entity_creation(
        etype=etype,
        worker_id=task_distributor.process_index,
        worker_cnt=task_distributor.num_processes,
    )
    task_distributor.push_new_tasks(tasks)



"""
Platform config file reader and config model.
"""

import os
from typing import Annotated, Optional, Union

import yaml
from pydantic import (
    BaseModel,
    ConfigDict,
    Extra,
    Field,
    NonNegativeInt,
    PositiveInt,
    field_validator,
    model_validator,
)
from pydantic_core.core_schema import FieldValidationInfo

from dp3.common.attrspec import (
    AttrSpec,
    AttrSpecClassic,
    AttrSpecGeneric,
    AttrSpecReadOnly,
    AttrSpecType,
)
from dp3.common.entityspec import EntitySpec


class NoDefault:
    pass


class MissingConfigError(Exception):
    pass


class HierarchicalDict(dict):
    """Extension of built-in `dict` that simplifies working with a nested hierarchy of dicts."""

    def __repr__(self):
        return f"HierarchicalDict({dict.__repr__(self)})"

    def copy(self):
        return HierarchicalDict(dict.copy(self))

    def get(self, key, default=NoDefault):
        """
        Key may be a path (in dot notation) into a hierarchy of dicts. For example
          `dictionary.get('abc.x.y')`
        is equivalent to
          `dictionary['abc']['x']['y']`.

        :returns: `self[key]` or `default` if key is not found.
        """
        d = self
        try:
            while "." in key:
                first_key, key = key.split(".", 1)
                d = d[first_key]
            return d[key]
        except (KeyError, TypeError):
            pass  # not found - continue below
        if default is NoDefault:
            raise MissingConfigError("Mandatory configuration element is missing: " + key)
        else:
            return default

    def update(self, other, **kwargs):
        """
        Update `HierarchicalDict` with other dictionary and merge common keys.

        If there is a key in both current and the other dictionary and values of
        both keys are dictionaries, they are merged together.

        Example:
        ```
        HierarchicalDict({'a': {'b': 1, 'c': 2}}).update({'a': {'b': 10, 'd': 3}})
        ->
        HierarchicalDict({'a': {'b': 10, 'c': 2, 'd': 3}})
        ```
        Changes the dictionary directly, returns `None`.
        """
        other = dict(other)
        for key in other:
            if key in self:
                if isinstance(self[key], dict) and isinstance(other[key], dict):
                    # The key is present in both dicts and both key values are dicts -> merge them
                    HierarchicalDict.update(self[key], other[key])
                else:
                    # One of the key values is not a dict -> overwrite the value
                    # in self by the one from other (like normal "update" does)
                    self[key] = other[key]
            else:
                # key is not present in self -> set it to value from other
                self[key] = other[key]


def read_config(filepath: str) -> HierarchicalDict:
    """
    Read configuration file and return config as a dict-like object.

    The configuration file should contain a valid YAML
    - Comments may be included as lines starting with `#` (optionally preceded
      by whitespaces).

    This function reads the file and converts it to a `HierarchicalDict`.
    The only difference from built-in `dict` is its `get` method, which allows
    hierarchical keys (e.g. `abc.x.y`).
    See [doc of get method][dp3.common.config.HierarchicalDict.get] for more information.
    """
    with open(filepath) as file_content:
        return HierarchicalDict(yaml.safe_load(file_content))


def read_config_dir(dir_path: str, recursive: bool = False) -> HierarchicalDict:
    """
    Same as [read_config][dp3.common.config.read_config],
    but it loads whole configuration directory of YAML files,
    so only files ending with ".yml" are loaded.
    Each loaded configuration is located under key named after configuration filename.

    Args:
        dir_path: Path to read config from.
        recursive: If `recursive` is set, then the configuration directory will be read
            recursively (including configuration files inside directories).
    """
    all_files_paths = os.listdir(dir_path)
    config = HierarchicalDict()
    for config_filename in all_files_paths:
        config_full_path = os.path.join(dir_path, config_filename)
        if os.path.isdir(config_full_path) and recursive:
            loaded_config = read_config_dir(config_full_path, recursive)
        elif os.path.isfile(config_full_path) and config_filename.endswith(".yml"):
            try:
                loaded_config = read_config(config_full_path)
            except TypeError:
                # configuration file is empty
                continue
            # remove '.yml' suffix of filename
            config_filename = config_filename[:-4]
        else:
            continue
        # place configuration files into another dictionary level named by config dictionary name
        config[config_filename] = loaded_config
    return config


CRON_EXPR_PATTERN = r"^((\d+,)+\d+|(\d+-\d+)|\d+|(\*\/\d+)|\*)$"
CronStr = Annotated[str, Field(pattern=CRON_EXPR_PATTERN)]
TimeInt = Annotated[int, Field(ge=0, le=59)]


class CronExpression(BaseModel, extra=Extra.forbid):
    """
    Cron expression used for scheduling. Also support standard cron expressions, such as

    - "*/15" (every 15 units)
    - "1,2,3" (1, 2 and 3)
    - "1-3" (1, 2 and 3)

    Attributes:
        year: 4-digit year
        month: month (1-12)
        day: day of month (1-31)
        week: ISO week (1-53)
        day_of_week: number or name of weekday (0-6 or mon,tue,wed,thu,fri,sat,sun)
        hour: hour (0-23)
        minute: minute (0-59)
        second: second (0-59)
        timezone: Timezone for time specification (default is UTC).
    """

    second: Optional[Union[TimeInt, CronStr]] = None
    minute: Optional[Union[TimeInt, CronStr]] = None
    hour: Optional[Union[TimeInt, CronStr]] = None

    day: Optional[Union[Annotated[int, Field(ge=1, le=31)], CronStr]] = None
    day_of_week: Optional[Union[Annotated[int, Field(ge=0, le=6)], CronStr]] = None

    week: Optional[int] = Field(default=None, ge=1, le=53)
    month: Optional[int] = Field(default=None, ge=1, le=12)
    year: Optional[str] = Field(default=None, pattern=r"^\d{4}$")

    timezone: str = "UTC"


class EntitySpecDict(BaseModel):
    """Class representing full specification of an entity.

    Attributes:
        entity: Specification and settings of entity itself.
        attribs: A mapping of attribute id -> AttrSpec
    """

    entity: EntitySpec
    attribs: dict[str, AttrSpecType]

    def __getitem__(self, item):
        return self.__getattribute__(item)

    @field_validator("entity", mode="before")
    def _parse_entity_spec(cls, v):
        if isinstance(v, EntitySpec):
            return v
        return EntitySpec.model_validate(v)

    @field_validator("attribs", mode="before")
    def _parse_attr_spec(cls, v):
        assert isinstance(v, dict), "'attribs' must be a dictionary"
        return {
            attr_id: AttrSpec(attr_id, spec) if not isinstance(spec, AttrSpecGeneric) else spec
            for attr_id, spec in v.items()
        }


class ModelSpec(BaseModel):
    """
    Class representing the platform's current entity and attribute specification.

    Attributes:
        config: Legacy config format, exactly mirrors the config files.

        entities: Mapping of entity id -> EntitySpec
        attributes: Mapping of (entity id, attribute id) -> AttrSpec
        entity_attributes: Mapping of entity id -> attribute id -> AttrSpec

        relations: Mapping of (entity id, attribute id) -> AttrSpec
            only contains attributes which are relations.
    """

    config: dict[str, EntitySpecDict]

    entities: dict[str, EntitySpec]
    attributes: dict[tuple[str, str], AttrSpecType]
    entity_attributes: dict[str, dict[str, AttrSpecType]]

    relations: dict[tuple[str, str], AttrSpecType]

    def __init__(self, config: HierarchicalDict):
        """
        Provided configuration must be a dict of following structure:
        ```
        {
            <entity type>: {
                'entity': {
                    entity specification
                },
                'attribs': {
                    <attr id>: {
                        attribute specification
                    },
                    other attributes
                }
            },
            other entity types
        }
        ```
        Raises:
            ValueError: if the specification is invalid.
        """
        super().__init__(
            config=config, entities={}, attributes={}, entity_attributes={}, relations={}
        )

    @field_validator("entities")
    def _fill_entities(cls, v, info: FieldValidationInfo):
        if "config" not in info.data:
            return v
        return {
            entity_id: entity_dict["entity"]
            for entity_id, entity_dict in info.data["config"].items()
        }

    @field_validator("attributes")
    def _fill_attributes(cls, v, info: FieldValidationInfo):
        if "config" not in info.data:
            return v
        return {
            (entity_id, attr_id): attr_spec
            for entity_id, entity_dict in info.data["config"].items()
            for attr_id, attr_spec in entity_dict["attribs"].items()
        }

    @field_validator("entity_attributes")
    def _fill_entity_attributes(cls, v, info: FieldValidationInfo):
        if "config" not in info.data:
            return v
        return {
            entity_id: dict(entity_dict["attribs"].items())
            for entity_id, entity_dict in info.data["config"].items()
        }

    @field_validator("relations")
    def _fill_relations(cls, v, info: FieldValidationInfo):
        if "attributes" not in info.data:
            return v
        return {
            entity_id_attr_id: attr_spec
            for entity_id_attr_id, attr_spec in info.data["attributes"].items()
            if isinstance(attr_spec, AttrSpecClassic) and attr_spec.is_relation
        }

    @model_validator(mode="after")
    def _validate_relations(self):
        """Validate that relation type attributes link to existing entities."""
        for entity_attr, attr_spec in self.relations.items():
            if attr_spec.relation_to not in self.entities:
                entity, attr = entity_attr
                raise ValueError(
                    f"'{attr_spec.relation_to}', linked by '{attr}' is not a valid entity."
                )
        return self

    @model_validator(mode="after")
    def _fill_and_validate_mirrors(self):
        """Validate that relation mirrors do not reference existing attributes and create them."""
        for (entity, attr), attr_spec in self.relations.items():
            if not attr_spec.is_mirrored:
                continue

            linked_entity = attr_spec.relation_to
            linked_attr = attr_spec.mirror_as

            mirror_attr = AttrSpecReadOnly(
                id=linked_attr,
                name=linked_attr,
                data_type=f"set<link<{entity}>>",
                type="plain",
                description=f"Read-only mirror attribute of {entity}.{attr}",
            )

            # We will accept correct definitions of mirrored attributes to fix errors when
            # a `ModelSpec` instance is validated again (e.g. when passed to `PlatformConfig`)
            # It's not pretty, but better than requiring a context like in `DataPointTask`.
            configured = self.attributes.get((linked_entity, linked_attr))
            if configured and configured.model_dump() != mirror_attr.model_dump():
                raise ValueError(
                    f"'{linked_entity}.{linked_attr}' is a mirrored attribute, "
                    "but already exists in configuration. "
                    "Mirrored attributes are defined implicitly, remove the definition."
                )

            self.config[linked_entity]["attribs"][linked_attr] = mirror_attr
            self.attributes[linked_entity, linked_attr] = mirror_attr
            self.entity_attributes[linked_entity][linked_attr] = mirror_attr

        return self

    def attr(self, entity_type: str, attr: str) -> AttrSpecType:
        return self.attributes[entity_type, attr]

    def attribs(self, entity_type: str) -> dict[str, AttrSpecType]:
        return self.entity_attributes[entity_type]

    def entity(self, entity_type: str) -> EntitySpec:
        return self.entities[entity_type]

    def items(self):
        return self.config.items()

    def keys(self):
        return self.config.keys()

    def __contains__(self, item):
        return item in self.config

    def __getitem__(self, item):
        return self.config[item]

    def __setitem__(self, key, value):
        self.config[key] = value

    def __repr__(self):
        return f"ModelSpec({','.join(self.config.keys())})"


class PlatformConfig(BaseModel):
    """
    An aggregation of configuration available to modules.

    Attributes:
        app_name: Name of the application, used when naming various structures of the platform
        config_base_path: Path to directory containing platform config
        config: A dictionary that contains the platform config
        model_spec: Specification of the platform's model (entities and attributes)

        num_processes: Number of worker processes
        process_index: Index of current process
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, protected_namespaces=())

    app_name: str
    config_base_path: str
    config: HierarchicalDict
    model_spec: ModelSpec

    num_processes: PositiveInt
    process_index: NonNegativeInt

    @field_validator("process_index")
    def valid_process_index(cls, v, info: FieldValidationInfo):
        if "num_processes" not in info.data:
            return v

        assert (
            v < info.data["num_processes"]
        ), "Process index must be less than total number of processes"
        return v

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return (
            f"PlatformConfig({self.app_name},"
            f" {self.config_base_path}, "
            # f" {repr(self.config)}, "
            f" {repr(self.model_spec)}, "
            f"num_processes={self.num_processes}, "
            f"process_index={self.process_index})"
        )
